@@ -1,19 +1,22 @@
-#include <sirius.h>
+#include <sirius.hpp>
 
 using namespace sirius;
 
 void test_comm_split(int comm_size)
 {
-    if (mpi_comm_world().rank() == 0) {
+    if (Communicator::world().rank() == 0) {
         printf("sub comm size: %i\n", comm_size);
     }
 
-    auto c1 = mpi_comm_world().split(mpi_comm_world().rank() / comm_size);
-    auto c2 = mpi_comm_world().split(mpi_comm_world().rank() % comm_size);
+    auto c1 = Communicator::world().split(Communicator::world().rank() / comm_size);
+    auto c2 = Communicator::world().split(Communicator::world().rank() % comm_size);
 
-    runtime::pstdout pout(mpi_comm_world());
+    auto c3 = c1;
+    Communicator c4(c2);
+
+    pstdout pout(Communicator::world());
     
-    pout.printf("global rank: %i, c1.rank: %i, c2.rank: %i\n", mpi_comm_world().rank(), c1.rank(), c2.rank());
+    pout.printf("global rank: %i, c1.rank: %i, c2.rank: %i\n", Communicator::world().rank(), c1.rank(), c2.rank());
 }
 
 int main(int argn, char** argv)
