@@ -1,4 +1,4 @@
-#include <sirius.h>
+#include <sirius.hpp>
 
 using namespace sirius;
 
@@ -9,6 +9,9 @@ void test_wf_inner(std::vector<int> mpi_grid_dims__,
                    linalg_t la__,
                    memory_t mem__)
 {
+    spla::Context spla_ctx(
+        la__ == linalg_t::blas || la__ == linalg_t::lapack || la__ == linalg_t::scalapack ? SPLA_PU_HOST : SPLA_PU_GPU);
+
     std::unique_ptr<BLACS_grid> blacs_grid;
     if (mpi_grid_dims__[0] * mpi_grid_dims__[1] == 1) {
         blacs_grid = std::unique_ptr<BLACS_grid>(new BLACS_grid(Communicator::self(), mpi_grid_dims__[0], mpi_grid_dims__[1]));
@@ -30,7 +33,7 @@ void test_wf_inner(std::vector<int> mpi_grid_dims__,
     }
 
     int nsp{1};
-    Wave_functions phi(gvp, num_bands__, memory_t::host, nsp);
+    Wave_functions<double> phi(gvp, num_bands__, memory_t::host, nsp);
 
     for (int is = 0; is < nsp; is++) {
         phi.zero(device_t::CPU, is, 0, num_bands__);
@@ -54,13 +57,13 @@ void test_wf_inner(std::vector<int> mpi_grid_dims__,
     }
 
     /* warmup call */
-    inner(mem__, la__, 0, phi, 0, num_bands__, phi, 0, num_bands__, ovlp, 0, 0);
+    inner(spla_ctx, spin_range(0), phi, 0, num_bands__, phi, 0, num_bands__, ovlp, 0, 0);
     Communicator::world().barrier();
 
-    utils::timer t1("inner");
-    inner(mem__, la__, 0, phi, 0, num_bands__, phi, 0, num_bands__, ovlp, 0, 0);
+    double t = -utils::wtime();
+    inner(spla_ctx, spin_range(0), phi, 0, num_bands__, phi, 0, num_bands__, ovlp, 0, 0);
     Communicator::world().barrier();
-    double t = t1.stop();
+    t += utils::wtime();
 
     double perf = 8e-9 * num_bands__ * num_bands__ *  gvec.num_gvec() / t;
     if (Communicator::world().rank() == 0) {
@@ -102,7 +105,7 @@ int main(int argn, char** argv)
         args.print_help();
         return 0;
     }
-    auto mpi_grid_dims = args.value<std::vector<int>>("mpi_grid_dims", {1, 1});
+    auto mpi_grid_dims = args.value("mpi_grid_dims", std::vector<int>({1, 1}));
     auto cutoff = args.value<double>("cutoff", 8.0);
     auto bs = args.value<int>("bs", 32);
     auto num_bands = args.value<int>("num_bands", 100);
@@ -114,8 +117,14 @@ int main(int argn, char** argv)
     test_wf_inner(mpi_grid_dims, cutoff, num_bands, bs, get_linalg_t(linalg_t_str), get_memory_t(memory_t_str));
 
     Communicator::world().barrier();
-    if (Communicator::world().rank() == 0) {
-        utils::timer::print();
+    int my_rank = Communicator::world().rank();
+
+    sirius::finalize(1);
+
+    if (my_rank == 0)  {
+        const auto timing_result = ::utils::global_rtgraph_timer.process();
+        std::cout << timing_result.print();
+        //std::ofstream ofs("timers.json", std::ofstream::out | std::ofstream::trunc);
+        //ofs << timing_result.json();
     }
-    sirius::finalize();
 }
