@@ -1,24 +1,24 @@
-#include <sirius.h>
+#include <sirius.hpp>
 
 using namespace sirius;
 
-double vlen(vector3d<double> v)
-{
-    return v.length();
-}
-
 int run_test()
 {
-    #pragma omp parallel 
-    {
-    vector3d<double> a(1, 2, 3);
+    vector3d<double> a(1.1, 2.2, 3.3);
     vector3d<double> b = a;
     vector3d<double> c(b);
     vector3d<double> d;
     d = c;
-    matrix3d<double> R = {{1,0,0},{0,1,0},{0,0,1}};
-    auto e = R * a + b;
-    printf("%f\n", vlen(e));
+    matrix3d<int> R = {{1,0,0},{0,1,0},{0,0,1}};
+    auto e = dot(R, a) + b;
+    vector3d<double> ref(2.2, 4.4, 6.6);
+    if ((ref - e).length() > 1e-16) {
+        return 1;
+    }
+    auto x = a + vector3d<int>(4, 5, 6);
+    vector3d<double> ref1(5.1, 7.2, 9.3);
+    if ((ref1 - x).length() > 1e-16) {
+        return 2;
     }
     return 0;
 }
