@@ -1,15 +1,19 @@
 from .helpers import *
-from .coefficient_array import *
+from .coefficient_array import CoefficientArray, PwCoeffs
 from .py_sirius import *
-from .py_sirius import K_point_set
+from .py_sirius import smearing as smearing
+from .py_sirius import K_point_set, Density
 from .logger import Logger
 from .operators import S_operator
 import numpy as np
 from numpy import array, zeros
-__all__ = ["ot", "baarman", "bands", "edft"]
+__all__ = ["ot", "baarman", "bands", "edft", "smearing"]
 
 
 class OccupancyDescriptor(object):
+    """
+    Accessor for occupation numbers
+    """
     def __set__(self, instance, value):
         for key, v in value._data.items():
             k, ispn = key
@@ -22,7 +26,6 @@ class OccupancyDescriptor(object):
         instance.sync_band_occupancies()
 
     def __get__(self, instance, owner):
-        import numpy as np
 
         out = CoefficientArray(dtype=np.double, ctype=np.array)
 
@@ -34,6 +37,9 @@ class OccupancyDescriptor(object):
 
 
 class PWDescriptor(object):
+    """
+    Accessor for wave-function coefficients
+    """
     def __set__(self, instance, value):
         from .helpers import store_pw_coeffs
         store_pw_coeffs(instance, value)
@@ -43,6 +49,9 @@ class PWDescriptor(object):
 
 
 class KPointWeightDescriptor(object):
+    """
+    Accessor for k-point weights
+    """
     def __get__(self, instance, owner):
 
         out = CoefficientArray(dtype=np.double, ctype=np.array)
@@ -55,8 +64,10 @@ class KPointWeightDescriptor(object):
 
 
 class BandEnergiesDescriptor(object):
+    """
+    Accessor for band energies
+    """
     def __get__(self, instance, owner):
-        from .coefficient_array import CoefficientArray
 
         out = CoefficientArray(dtype=np.double, ctype=np.array)
 
@@ -66,6 +77,7 @@ class BandEnergiesDescriptor(object):
                 out[key] = np.array(instance[k].band_energies(ispn))
         return out
 
+
     def __set__(self, instance, value):
         for key, val in value._data.items():
             k, ispn = key
@@ -73,7 +85,24 @@ class BandEnergiesDescriptor(object):
                 instance[k].set_band_energy(j, ispn, v)
         instance.sync_band_energies()
 
+
+class DensityDescriptor(object):
+    def __init__(self, i):
+        self.i = i
+
+    def __get__(self, instance, owner):
+        return np.array(instance.f_pw_local(self.i))
+
+    def __set__(self, instance, value):
+        instance.f_pw_local(self.i)[:] = value
+
+
 K_point_set.fn = OccupancyDescriptor()
 K_point_set.C = PWDescriptor()
 K_point_set.w = KPointWeightDescriptor()
 K_point_set.e = BandEnergiesDescriptor()
+
+Density.rho = DensityDescriptor(0)
+Density.mx = DensityDescriptor(1)
+Density.my = DensityDescriptor(2)
+Density.mz = DensityDescriptor(3)
