@@ -1,4 +1,4 @@
-from ..py_sirius import ewald_energy, energy_bxc, Wave_functions, MemoryEnum
+from ..py_sirius import ewald_energy, energy_bxc, Wave_functions, MemoryEnum, Hamiltonian0
 from ..coefficient_array import PwCoeffs
 import numpy as np
 
@@ -20,8 +20,7 @@ def pp_total_energy(potential, density, k_point_set, ctx):
     # TODO: Ewald energy is constant...
     return (k_point_set.valence_eval_sum() - potential.energy_vxc(density) -
             potential.PAW_one_elec_energy() - 0.5 * potential.energy_vha() -
-            energy_bxc(density, potential,
-                       ctx.num_mag_dims()) + potential.energy_exc(density) +
+            energy_bxc(density, potential) + potential.energy_exc(density) +
             potential.PAW_total_energy() + ewald_energy(ctx, gvec, unit_cell))
 
 
@@ -44,30 +43,28 @@ class Energy:
         else:
             self.ctx = ctx
 
-    def compute(self, cn):
+    def compute(self, X):
         """
         Keyword Arguments:
-        cn  --
-        ek  -- band energies
+        X  -- PW coefficients
+
+        Returns
+        Etot -- total energy
+        HX   -- Hamiltonian@X
         """
+        assert isinstance(X, PwCoeffs)
 
-        assert isinstance(cn, PwCoeffs)
-
-        for key, val in cn.items():
+        for key, val in X.items():
             k, ispn = key
             self.kpointset[k].spinor_wave_functions().pw_coeffs(ispn)[:, :val.shape[1]] = val
         # copy to device (if needed)
-        for ki in cn.kvalues():
+        for ki in X.kvalues():
             psi = self.kpointset[ki].spinor_wave_functions()
             if psi.preferred_memory_t() == MemoryEnum.device:
                 psi.copy_to_gpu()
         # update density, potential
-        self.density.generate(self.kpointset)
-        if self.ctx.use_symmetry():
-            self.density.symmetrize()
-            self.density.symmetrize_density_matrix()
+        self.density.generate(self.kpointset, self.ctx.use_symmetry())
 
-        self.density.generate_paw_loc_density()
         self.density.fft_transform(1)
         self.potential.generate(self.density)
         if self.ctx.use_symmetry():
@@ -75,13 +72,14 @@ class Energy:
 
         self.potential.fft_transform(1)
 
-        yn = self.H(cn, scale=False)
+        yn = self.H(X, scale=False)
+
         for key, val in yn.items():
             k, ispn = key
             benergies = np.zeros(self.ctx.num_bands(), dtype=np.complex)
             benergies[:val.shape[1]] = np.einsum('ij,ij->j',
-                                                val,
-                                                np.conj(cn[key]))
+                                                 val,
+                                                 np.conj(X[key]))
 
             for j, ek in enumerate(benergies):
                 self.kpointset[k].set_band_energy(j, ispn, np.real(ek))
@@ -99,9 +97,9 @@ class Energy:
 
 
 class ApplyHamiltonian:
-    def __init__(self, hamiltonian, kpointset):
-        assert (not isinstance(hamiltonian, ApplyHamiltonian))
-        self.hamiltonian = hamiltonian
+    def __init__(self, potential, kpointset):
+        assert not isinstance(potential, ApplyHamiltonian)
+        self.potential = potential
         self.kpointset = kpointset
 
     def apply(self, cn, scale=True, ki=None, ispn=None):
@@ -110,9 +108,11 @@ class ApplyHamiltonian:
         cn -- input coefficient array
         """
         from ..coefficient_array import PwCoeffs
-        ctx = self.hamiltonian.ctx()
+        from ..py_sirius import apply_hamiltonian
+
+        ctx = self.kpointset.ctx()
         num_sc = ctx.num_spins()
-        self.hamiltonian._apply_ref_inner_prepare()
+        H0 = Hamiltonian0(self.potential)
         if isinstance(cn, PwCoeffs):
             assert (ki is None)
             assert (ispn is None)
@@ -129,7 +129,7 @@ class ApplyHamiltonian:
                                        num_sc)
                 for i, val in ispn_coeffs:
                     Psi_x.pw_coeffs(i)[:, :val.shape[1]] = val
-                self.hamiltonian._apply_ref_inner(self.kpointset[k], Psi_y, Psi_x)
+                apply_hamiltonian(H0, kpoint, Psi_y, Psi_x)
 
                 w = kpoint.weight()
                 # copy coefficients from Psi_y
@@ -142,28 +142,7 @@ class ApplyHamiltonian:
                     else:
                         out[(k, i)] = np.array(
                             Psi_y.pw_coeffs(i), copy=False)[:, :num_wf]
-                # end for
-            self.hamiltonian._apply_ref_inner_dismiss()
             return out
-        else:
-            assert (ki in self.kpointset)
-            kpoint = self.kpointset[ki]
-            w = kpoint.weight()
-            num_wf = cn.shape[1]
-            Psi_y = Wave_functions(kpoint.gkvec_partition(), num_wf, ctx.preferred_memory_t(), num_sc)
-            Psi_x = Wave_functions(kpoint.gkvec_partition(), num_wf, ctx.preferred_memory_t(), num_sc)
-            bnd_occ = np.array(kpoint.band_occupancy(ispn))
-            Psi_x.pw_coeffs(ispn)[:] = cn
-            self.hamiltonian.apply_ref(kpoint, Psi_y, Psi_x)
-            self.hamiltonian._apply_ref_inner_dismiss()
-            if scale:
-                return np.matrix(
-                    np.array(Psi_y.pw_coeffs(ispn), copy=False) * bnd_occ * w,
-                    copy=True)
-            else:
-                return np.matrix(
-                    np.array(Psi_y.pw_coeffs(ispn), copy=False),
-                    copy=True)
 
     def __matmul__(self, cn):
         """
