@@ -19,20 +19,20 @@
 
 /** \file dmatrix.hpp
  *
- *  \brief Contains definition and implementaiton of sddk::dmatrix class.
+ *  \brief Contains definition and implementation of sddk::dmatrix class.
  */
 
 #ifndef __DMATRIX_HPP__
 #define __DMATRIX_HPP__
 
-#include "blacs_grid.hpp"
+#include "linalg/blacs_grid.hpp"
 #include "splindex.hpp"
 #include "hdf5_tree.hpp"
+#include <spla/spla.hpp>
+#include "type_definition.hpp"
 
 namespace sddk {
 
-using double_complex = std::complex<double>;
-
 /// Distributed matrix.
 template <typename T>
 class dmatrix : public matrix<T>
@@ -62,9 +62,12 @@ class dmatrix : public matrix<T>
     /// ScaLAPACK matrix descriptor.
     ftn_int descriptor_[9];
 
+    /// matrix distribution used for SPLA library functions
+    spla::MatrixDistribution spla_distri_ = spla::MatrixDistribution::create_mirror(MPI_COMM_SELF);
+
     void init()
     {
-#ifdef __SCALAPACK
+#ifdef SIRIUS_SCALAPACK
         if (blacs_grid_ != nullptr) {
             linalg_base::descinit(descriptor_, num_rows_, num_cols_, bs_row_, bs_col_, 0, 0, blacs_grid_->context(),
                                   spl_row_.local_size());
@@ -74,7 +77,7 @@ class dmatrix : public matrix<T>
 
     /* forbid copy constructor */
     dmatrix(dmatrix<T> const& src) = delete;
-    /* forbid assigment operator */
+    /* forbid assignment operator */
     dmatrix<T>& operator=(dmatrix<T> const& src) = delete;
 
   public:
@@ -83,34 +86,19 @@ class dmatrix : public matrix<T>
     {
     }
 
-    dmatrix(int num_rows__,
-            int num_cols__,
-            BLACS_grid const& blacs_grid__,
-            int bs_row__,
-            int bs_col__,
+    dmatrix(int num_rows__, int num_cols__, BLACS_grid const& blacs_grid__, int bs_row__, int bs_col__,
             memory_t mem_type__ = memory_t::host);
 
-    dmatrix(int num_rows__,
-            int num_cols__,
-            memory_t mem_type__ = memory_t::host);
+    dmatrix(int num_rows__, int num_cols__, memory_t mem_type__ = memory_t::host);
 
-    dmatrix(T* ptr__,
-            int num_rows__,
-            int num_cols__,
-            BLACS_grid const& blacs_grid__,
-            int bs_row__,
-            int bs_col__);
+    dmatrix(int num_rows__, int num_cols__, memory_pool& mp__, std::string const& label__ = "");
 
-    dmatrix(memory_pool& mp__,
-            int num_rows__,
-            int num_cols__,
-            BLACS_grid const& blacs_grid__,
-            int bs_row__,
-            int bs_col__);
+    dmatrix(T* ptr__, int num_rows__, int num_cols__, BLACS_grid const& blacs_grid__, int bs_row__, int bs_col__);
 
-    dmatrix(T* ptr__,
-            int num_rows__,
-            int num_cols__);
+    dmatrix(int num_rows__, int num_cols__, BLACS_grid const& blacs_grid__, int bs_row__, int bs_col__,
+            memory_pool& mp__);
+
+    dmatrix(T* ptr__, int num_rows__, int num_cols__);
 
     dmatrix(dmatrix<T>&& src) = default;
 
@@ -125,6 +113,11 @@ class dmatrix : public matrix<T>
         return -1;
     }
 
+    inline int size_local() const
+    {
+      return this->num_rows_local() * this->num_cols_local();
+    }
+
     /// Return number of rows in the global matrix.
     inline int num_rows() const
     {
@@ -176,6 +169,11 @@ class dmatrix : public matrix<T>
         return descriptor_;
     }
 
+    inline spla::MatrixDistribution& spla_distribution()
+    {
+        return spla_distri_;
+    }
+
     //void zero(int ir0__, int ic0__, int nr__, int nc__)
     //{
     //    splindex<splindex_t::block_cyclic> spl_r0(ir0__, blacs_grid().num_ranks_row(), blacs_grid().rank_row(), bs_row_);
@@ -203,7 +201,7 @@ class dmatrix : public matrix<T>
 
     void add(const int irow_glob, const int icol_glob, T val);
 
-    void add(double beta__, const int irow_glob, const int icol_glob, T val);
+    void add(real_type<T> beta__, const int irow_glob, const int icol_glob, T val);
 
     void make_real_diag(int n__);
 
@@ -257,7 +255,107 @@ class dmatrix : public matrix<T>
 
     void save_to_hdf5(std::string name__, int m__, int n__);
 
-    void serialize(std::string name__, int n__) const;
+    template <typename F = T, std::enable_if_t<!std::is_scalar<F>::value, bool> = true>
+    void serialize(std::string name__, int n__) const
+    {
+        mdarray<T, 2> full_mtrx(num_rows(), num_cols());
+        full_mtrx.zero();
+
+        for (int j = 0; j < num_cols_local(); j++) {
+            for (int i = 0; i < num_rows_local(); i++) {
+                full_mtrx(irow(i), icol(j)) = (*this)(i, j);
+            }
+        }
+        if (blacs_grid_) {
+            blacs_grid_->comm().allreduce(full_mtrx.at(memory_t::host), static_cast<int>(full_mtrx.size()));
+        }
+
+        // json dict;
+        // dict["mtrx_re"] = json::array();
+        // for (int i = 0; i < num_rows(); i++) {
+        //    dict["mtrx_re"].push_back(json::array());
+        //    for (int j = 0; j < num_cols(); j++) {
+        //        dict["mtrx_re"][i].push_back(full_mtrx(i, j).real());
+        //    }
+        //}
+
+        if (!blacs_grid_ || blacs_grid_->comm().rank() == 0) {
+            // std::cout << "mtrx: " << name__ << std::endl;
+            // std::cout << dict.dump(4);
+
+            std::printf("matrix label: %s\n", name__.c_str());
+            std::printf("{\n");
+            for (int i = 0; i < n__; i++) {
+                std::printf("{");
+                for (int j = 0; j < n__; j++) {
+                    std::printf("%18.13f + I * %18.13f", full_mtrx(i, j).real(), full_mtrx(i, j).imag());
+                    if (j != n__ - 1) {
+                        std::printf(",");
+                    }
+                }
+                if (i != n__ - 1) {
+                    std::printf("},\n");
+                } else {
+                    std::printf("}\n");
+                }
+            }
+            std::printf("}\n");
+        }
+
+        // std::ofstream ofs(aiida_output_file, std::ofstream::out | std::ofstream::trunc);
+        // ofs << dict.dump(4);
+    }
+
+    template <typename F = T, std::enable_if_t<std::is_scalar<F>::value, bool> = true>
+    void serialize(std::string name__, int n__) const
+    {
+        mdarray<T, 2> full_mtrx(num_rows(), num_cols());
+        full_mtrx.zero();
+
+        for (int j = 0; j < num_cols_local(); j++) {
+            for (int i = 0; i < num_rows_local(); i++) {
+                full_mtrx(irow(i), icol(j)) = (*this)(i, j);
+            }
+        }
+        blacs_grid_->comm().allreduce(full_mtrx.at(memory_t::host), static_cast<int>(full_mtrx.size()));
+
+        // json dict;
+        // dict["mtrx"] = json::array();
+        // for (int i = 0; i < num_rows(); i++) {
+        //    dict["mtrx"].push_back(json::array());
+        //    for (int j = 0; j < num_cols(); j++) {
+        //        dict["mtrx"][i].push_back(full_mtrx(i, j));
+        //    }
+        //}
+
+        // if (blacs_grid_->comm().rank() == 0) {
+        //    std::cout << "mtrx: " << name__ << std::endl;
+        //    std::cout << dict.dump(4);
+        //}
+
+        // std::ofstream ofs(aiida_output_file, std::ofstream::out | std::ofstream::trunc);
+        // ofs << dict.dump(4);
+
+        if (blacs_grid_->comm().rank() == 0) {
+            std::printf("matrix label: %s\n", name__.c_str());
+            std::printf("{\n");
+            for (int i = 0; i < n__; i++) {
+                std::printf("{");
+                for (int j = 0; j < n__; j++) {
+                    std::printf("%18.13f", full_mtrx(i, j));
+                    if (j != n__ - 1) {
+                        std::printf(",");
+                    }
+                }
+                if (i != n__ - 1) {
+                    std::printf("},\n");
+                } else {
+                    std::printf("}\n");
+                }
+            }
+            std::printf("}\n");
+        }
+    }
 
     inline Communicator const& comm() const {
         if (blacs_grid_ != nullptr) {
@@ -266,6 +364,7 @@ class dmatrix : public matrix<T>
             return Communicator::self();
         }
     }
+
 };
 
 
