@@ -63,8 +63,17 @@ class vector3d : public std::array<T, 3>
         }
     }
 
+    vector3d& operator=(std::initializer_list<T> v__)
+    {
+        assert(v__.size() == 3);
+        for (int x : {0, 1, 2}) {
+            (*this)[x] = v__.begin()[x];
+        }
+        return *this;
+    }
+
     /// Create from std::vector.
-    vector3d(std::vector<T> v__)
+    vector3d(const std::vector<T>& v__)
     {
         assert(v__.size() == 3);
         for (int x : {0, 1, 2}) {
@@ -72,6 +81,15 @@ class vector3d : public std::array<T, 3>
         }
     }
 
+    vector3d& operator=(const std::vector<T>& v__)
+    {
+        assert(v__.size() == 3);
+        for (int x : {0, 1, 2}) {
+            (*this)[x] = v__[x];
+        }
+        return *this;
+    }
+
     /// Create from raw pointer.
     vector3d(T const* ptr__)
     {
@@ -114,26 +132,6 @@ class vector3d : public std::array<T, 3>
         return static_cast<double>(std::pow((*this)[0], 2) + std::pow((*this)[1], 2) + std::pow((*this)[2], 2));
     }
 
-    template <typename U>
-    inline vector3d<decltype(T{} + U{})> operator+(vector3d<U> const& b) const
-    {
-        vector3d<decltype(T{} + U{})> a = *this;
-        for (int x : {0, 1, 2}) {
-            a[x] += b[x];
-        }
-        return a;
-    }
-
-    template <typename U>
-    inline vector3d<decltype(T{} - U{})> operator-(vector3d<U> const& b) const
-    {
-        vector3d<decltype(T{} - U{})> a = *this;
-        for (int x : {0, 1, 2}) {
-            a[x] -= b[x];
-        }
-        return a;
-    }
-
     inline vector3d<T>& operator+=(vector3d<T> const& b)
     {
         for (int x : {0, 1, 2}) {
@@ -150,33 +148,59 @@ class vector3d : public std::array<T, 3>
         return *this;
     }
 
-    template <typename U>
-    inline friend vector3d<decltype(T{} * U{})> operator*(vector3d<T> vec, U p)
-    {
-        vector3d<decltype(T{} * U{})> a;
-        for (int x : {0, 1, 2}) {
-            a[x] = vec[x] * p;
-        }
-        return a;
-    }
-
-    template <typename U>
-    inline friend vector3d<decltype(T{} * U{})> operator*(U p, vector3d<T> vec)
-    {
-        return vec * p;
-    }
-
-    template <typename U>
-    inline friend vector3d<decltype(T{} * U{})> operator/(vector3d<T> vec, U p)
-    {
-        vector3d<decltype(T{} * U{})> a;
-        for (int x : {0, 1, 2}) {
-            a[x] = vec[x] / p;
-        }
-        return a;
-    }
 };
 
+template <typename T, typename U>
+inline vector3d<decltype(T{} + U{})>
+operator+(vector3d<T> const& a, vector3d<U> const& b)
+{
+    vector3d<decltype(T{} + U{})> c;
+    for (int x : {0, 1, 2}) {
+        c[x] = a[x] + b[x];
+    }
+    return c;
+}
+
+template <typename T, typename U>
+inline vector3d<decltype(T{} - U{})>
+operator-(vector3d<T> const& a, vector3d<U> const& b)
+{
+    vector3d<decltype(T{} - U{})> c;
+    for (int x : {0, 1, 2}) {
+        c[x] = a[x] - b[x];
+    }
+    return c;
+}
+
+template <typename T, typename U>
+inline std::enable_if_t<std::is_scalar<U>::value, vector3d<decltype(T{} * U{})>>
+operator*(vector3d<T> const& vec, U p)
+{
+    vector3d<decltype(T{} * U{})> a;
+    for (int x : {0, 1, 2}) {
+        a[x] = vec[x] * p;
+    }
+    return a;
+}
+
+template <typename T, typename U>
+inline std::enable_if_t<std::is_scalar<U>::value, vector3d<decltype(T{} * U{})>>
+operator*(U p, vector3d<T> const& vec)
+{
+    return vec * p;
+}
+
+template <typename T, typename U>
+inline std::enable_if_t<std::is_scalar<U>::value, vector3d<decltype(T{} * U{})>>
+operator/(vector3d<T> const& vec, U p)
+{
+    vector3d<decltype(T{} * U{})> a;
+    for (int x : {0, 1, 2}) {
+        a[x] = vec[x] / p;
+    }
+    return a;
+}
+
 template <typename T, typename U>
 inline auto dot(vector3d<T> const a, vector3d<U> const b) -> decltype(T{} * U{})
 {
@@ -196,7 +220,7 @@ inline vector3d<T> cross(vector3d<T> const a, vector3d<T> const b)
 template <typename T>
 std::ostream& operator<<(std::ostream& out, vector3d<T> const& v)
 {
-    out << v[0] << " " << v[1] << " " << v[2];
+    out << "{" << v[0] << ", " << v[1] << ", " << v[2] << "}";
     return out;
 }
 
@@ -234,9 +258,18 @@ class matrix3d
         }
     }
 
+    matrix3d(std::array<std::array<T, 3>, 3> src__)
+    {
+        for (int i = 0; i < 3; i++) {
+            for (int j = 0; j < 3; j++) {
+                mtrx_[i][j] = src__[i][j];
+            }
+        }
+    }
+
     /// Copy constructor.
     template <typename U>
-    matrix3d(const matrix3d<U>& src__)
+    matrix3d(matrix3d<U> const& src__)
     {
         for (int i = 0; i < 3; i++) {
             for (int j = 0; j < 3; j++) {
@@ -255,8 +288,8 @@ class matrix3d
         }
     }
 
-    /// Assigment operator.
-    matrix3d<T>& operator=(const matrix3d<T>& rhs)
+    /// Assignment operator.
+    matrix3d<T>& operator=(matrix3d<T> const& rhs)
     {
         if (this != &rhs) {
             std::memcpy(&this->mtrx_[0][0], &rhs.mtrx_[0][0], 9 * sizeof(T));
@@ -274,34 +307,6 @@ class matrix3d
         return mtrx_[i][j];
     }
 
-    /// Multiply two matrices.
-    template <typename U>
-    inline matrix3d<decltype(T{} * U{})> operator*(matrix3d<U> const& b) const
-    {
-        matrix3d<decltype(T{} * U{})> c;
-        for (int i = 0; i < 3; i++) {
-            for (int j = 0; j < 3; j++) {
-                for (int k = 0; k < 3; k++) {
-                    c(i, j) += (*this)(i, k) * b(k, j);
-                }
-            }
-        }
-        return c;
-    }
-
-    /// Matrix-vector multiplication.
-    template <typename U>
-    inline vector3d<decltype(T{} * U{})> operator*(vector3d<U> const& b) const
-    {
-        vector3d<decltype(T{} * U{})> a;
-        for (int i = 0; i < 3; i++) {
-            for (int j = 0; j < 3; j++) {
-                a[i] += (*this)(i, j) * b[j];
-            }
-        }
-        return a;
-    }
-
     /// Sum of two matrices.
     template <typename U>
     inline matrix3d<decltype(T{} + U{})> operator+(matrix3d<U> const& b) const
@@ -327,19 +332,6 @@ class matrix3d
         return *this;
     }
 
-    /// Multiply matrix by a scalar number.
-    template <typename U>
-    inline matrix3d<decltype(T{} * U{})> operator*(U p) const
-    {
-        matrix3d<decltype(T{} * U{})> c;
-        for (int i = 0; i < 3; i++) {
-            for (int j = 0; j < 3; j++) {
-                c(i, j) = (*this)(i, j) * p;
-            }
-        }
-        return c;
-    }
-
     /// Multiply matrix by a scalar number.
     template <typename U>
     inline matrix3d<T>& operator*=(U p)
@@ -366,6 +358,78 @@ class matrix3d
     }
 };
 
+/// Multiply matrix by a scalar number.
+template <typename T, typename U>
+inline std::enable_if_t<std::is_scalar<U>::value, matrix3d<decltype(T{} * U{})>> operator*(matrix3d<T> const& a__, U p__)
+{
+    matrix3d<decltype(T{} * U{})> c;
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < 3; j++) {
+            c(i, j) = a__(i, j) * p__;
+        }
+    }
+    return c;
+}
+
+template <typename T, typename U>
+inline std::enable_if_t<std::is_scalar<U>::value, matrix3d<decltype(T{} * U{})>> operator*(U p__, matrix3d<T> const& a__)
+{
+    return a__ * p__;
+}
+
+inline bool operator==(matrix3d<int> const& a__, matrix3d<int> const& b__)
+{
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < 3; j++) {
+            if (a__(i, j) != b__(i, j)) {
+                return false;
+            }
+        }
+    }
+    return true;
+}
+
+/// Multiply two matrices.
+template <typename T, typename U>
+inline matrix3d<decltype(T{} * U{})> dot(matrix3d<T> const& a__, matrix3d<U> const& b__)
+{
+    matrix3d<decltype(T{} * U{})> c;
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < 3; j++) {
+            for (int k = 0; k < 3; k++) {
+                c(i, j) += a__(i, k) * b__(k, j);
+            }
+        }
+    }
+    return c;
+}
+
+/// Matrix-vector multiplication.
+template <typename T, typename U>
+inline vector3d<decltype(T{} * U{})> dot(matrix3d<T> const& m__, vector3d<U> const& b__)
+{
+    vector3d<decltype(T{} * U{})> a;
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < 3; j++) {
+            a[i] += m__(i, j) * b__[j];
+        }
+    }
+    return a;
+}
+
+/// Vector-matrix multiplication.
+template <typename T, typename U>
+inline vector3d<decltype(T{} * U{})> dot(vector3d<U> const& b__, matrix3d<T> const& m__)
+{
+    vector3d<decltype(T{} * U{})> a;
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < 3; j++) {
+            a[i] += b__[j] * m__(j, i);
+        }
+    }
+    return a;
+}
+
 /// Return transpose of the matrix.
 template <typename T>
 inline matrix3d<T> transpose(matrix3d<T> src)
@@ -421,7 +485,7 @@ inline matrix3d<T> inverse(matrix3d<T> src)
 }
 
 template <typename T>
-inline std::ostream& operator<<(std::ostream& out, matrix3d<T>& v)
+inline std::ostream& operator<<(std::ostream& out, matrix3d<T> const& v)
 {
     out << "{";
     for (int i = 0; i < 3; i++) {
@@ -486,7 +550,7 @@ inline std::pair<vector3d<double>, vector3d<int>> reduce_coordinates(vector3d<do
 }
 
 /// Find supercell that circumscribes the sphere with a given radius.
-/** Serach for the translation limits (N1, N2, N3) such that the resulting supercell with the lattice
+/** Search for the translation limits (N1, N2, N3) such that the resulting supercell with the lattice
  *  vectors a1 * N1, a2 * N2, a3 * N3 fully contains the sphere with a given radius. This is done
  *  by equating the expressions for the volume of the supercell:
  *   Volume = |(A1 x A2) * A3| = N1 * N2 * N3 * |(a1 x a2) * a3|
@@ -508,6 +572,36 @@ inline vector3d<int> find_translations(double radius__, matrix3d<double> const&
     return {limits[0], limits[1], limits[2]};
 }
 
+/// Transform Cartesian coordinates [x,y,z] to spherical coordinates [r,theta,phi]
+inline vector3d<double> spherical_coordinates(vector3d<double> vc)
+{
+    geometry3d::vector3d<double> vs;
+
+    const double eps{1e-12};
+
+    const double twopi = 6.2831853071795864769;
+
+    vs[0] = vc.length();
+
+    if (vs[0] <= eps) {
+        vs[1] = 0.0;
+        vs[2] = 0.0;
+    } else {
+        vs[1] = std::acos(vc[2] / vs[0]); // theta = cos^{-1}(z/r)
+
+        if (std::abs(vc[0]) > eps || std::abs(vc[1]) > eps) {
+            vs[2] = std::atan2(vc[1], vc[0]); // phi = tan^{-1}(y/x)
+            if (vs[2] < 0.0) {
+                vs[2] += twopi;
+            }
+        } else {
+            vs[2] = 0.0;
+        }
+    }
+
+    return vs;
+}
+
 } // namespace geometry3d
 
 #endif // __GEOMETRY3D_HPP__
