@@ -19,10 +19,13 @@
 
 /** \file gvec.cpp
  *
- *  \brief Definitions.
+ *  \brief Contains the implementation of Gvec class.
  *
  */
+
+#include "symmetry/lattice.hpp"
 #include "gvec.hpp"
+#include "serializer.hpp"
 
 namespace sddk {
 
@@ -51,7 +54,7 @@ void sddk::Gvec::find_z_columns(double Gmax__, const FFT3D_grid& fft_box__)
         std::vector<int> zcol;
 
         /* in general case take z in [0, Nz) */
-        int zmax = fft_box__.size(2) - 1;
+        int zmax = fft_box__[2] - 1;
         /* in case of G-vector reduction take z in [0, Nz/2] for {x=0,y=0} stick */
         if (reduce_gvec_ && !i && !j) {
             zmax = fft_box__.limits(2).second;
@@ -61,7 +64,7 @@ void sddk::Gvec::find_z_columns(double Gmax__, const FFT3D_grid& fft_box__)
             /* get z-coordinate of G-vector */
             int k = fft_box__.freq_by_coord<2>(iz);
             /* take G+k */
-            auto vgk = lattice_vectors_ * (vector3d<double>(i, j, k) + vk_);
+            auto vgk = dot(lattice_vectors_, (vector3d<double>(i, j, k) + vk_));
             /* add z-coordinate of G-vector to the list */
             if (vgk.length() <= Gmax__) {
                 zcol.push_back(k);
@@ -94,7 +97,8 @@ void sddk::Gvec::find_z_columns(double Gmax__, const FFT3D_grid& fft_box__)
         }
     }
 
-    for (int i = fft_box__.limits(0).first; i <= fft_box__.limits(0).second; i++) {
+    // Check all z-columns and add if within sphere. Only allow non-negative x-indices for reduced case
+    for (int i = reduce_gvec_? 0 : fft_box__.limits(0).first; i <= fft_box__.limits(0).second; i++) {
         for (int j = fft_box__.limits(1).first; j <= fft_box__.limits(1).second; j++) {
             add_new_column(i, j);
         }
@@ -176,8 +180,10 @@ void Gvec::distribute_z_columns()
         ng += gvec_distr_.counts[rank];
     }
     if (ng != num_gvec_) {
-        TERMINATE("wrong number of G-vectors");
+        throw std::runtime_error("wrong number of G-vectors");
     }
+    this->offset_ = this->gvec_offset(this->comm().rank());
+    this->count_ = this->gvec_count(this->comm().rank());
 }
 
 void Gvec::find_gvec_shells()
@@ -191,7 +197,7 @@ void Gvec::find_gvec_shells()
     auto lat_sym = sirius::find_lat_sym(lattice_vectors_, 1e-6);
 
     num_gvec_shells_ = 0;
-    gvec_shell_      = mdarray<int, 1>(num_gvec_);
+    gvec_shell_      = sddk::mdarray<int, 1>(num_gvec_);
 
     std::fill(&gvec_shell_[0], &gvec_shell_[0] + num_gvec_, -1);
 
@@ -199,24 +205,23 @@ void Gvec::find_gvec_shells()
     for (int ig = 0; ig < num_gvec_; ig++) {
         if (gvec_shell_[ig] == -1) {
             auto G = gvec(ig);
-            for (size_t isym = 0; isym < lat_sym.size(); isym++) {
-                auto R   = lat_sym[isym];
-                auto G1  = R * G;
+            for (auto& R: lat_sym) {
+                auto G1  = dot(R, G);
                 auto ig1 = index_by_gvec(G1);
                 if (ig1 == -1) {
-                    auto G1 = R * (G * (-1));
-                    ig1     = index_by_gvec(G1);
-                }
-                if (ig1 >= 0) {
-                    gvec_shell_[ig1] = num_gvec_shells_;
+                    ig1 = index_by_gvec(G1 * (-1));
+                    if (ig1 == -1) {
+                        throw std::runtime_error("[sddk::Gvec] symmetry-related G-vector is not found");
+                    }
                 }
+                gvec_shell_[ig1] = num_gvec_shells_;
             }
             num_gvec_shells_++;
         }
     }
     for (int ig = 0; ig < num_gvec_; ig++) {
         if (gvec_shell_[ig] == -1) {
-            TERMINATE("wrong G-vector shell");
+            throw std::runtime_error("[sddk::Gvec] wrong G-vector shell");
         }
     }
 
@@ -229,25 +234,26 @@ void Gvec::find_gvec_shells()
         if (gvec_shell_len_[igsh] < 0) {
             gvec_shell_len_[igsh] = g;
         } else {
-            /* lattice symmetries were found wih 1e-6 tolererance for the metric tensor,
+            /* lattice symmetries were found with 1e-6 tolererance for the metric tensor,
                so tolerance on length should be square root of that */
             if (std::abs(gvec_shell_len_[igsh] - g) > 1e-3) {
                 std::stringstream s;
-                s << "wrong G-vector length"
-                  << "\n"
+                s << "[sddk::Gvec] wrong G-vector length\n"
                   << "  length of G-shell : " << gvec_shell_len_[igsh] << "\n"
                   << "  length of current G-vector: " << g << "\n"
                   << "  index of G-vector: " << ig << "\n"
                   << "  index of G-shell: " << igsh << "\n"
                   << "  length difference: " << std::abs(gvec_shell_len_[igsh] - g);
-                TERMINATE(s);
+                throw std::runtime_error(s.str());
             }
         }
     }
 
+    // TODO: maybe, make an average G-shell length.
+
     /* list of pairs (length, index of G-vector) */
     std::vector<std::pair<uint64_t, int>> tmp(num_gvec_);
-#pragma omp parallel for schedule(static)
+    #pragma omp parallel for schedule(static)
     for (int ig = 0; ig < num_gvec(); ig++) {
         /* make some reasonable roundoff */
         uint64_t len = static_cast<uint64_t>(gvec_shell_len_[gvec_shell_[ig]] * 1e10);
@@ -276,6 +282,28 @@ void Gvec::find_gvec_shells()
     }
     gvec_shell_len_ = mdarray<double, 1>(num_gvec_shells_);
     std::copy(tmp_len.begin(), tmp_len.end(), gvec_shell_len_.at(memory_t::host));
+
+    /* map from global index of G-shell to a list of local G-vectors */
+    std::map<int, std::vector<int>> gshmap;
+    for (int igloc = 0; igloc < this->count(); igloc++) {
+        int igsh = this->shell(this->offset() + igloc);
+        if (gshmap.count(igsh) == 0) {
+            gshmap[igsh] = std::vector<int>();
+        }
+        gshmap[igsh].push_back(igloc);
+    }
+
+    num_gvec_shells_local_ = 0;
+    gvec_shell_idx_local_.resize(this->count());
+    gvec_shell_len_local_.clear();
+    for (auto it = gshmap.begin(); it != gshmap.end(); ++it) {
+        int igsh = it->first;
+        gvec_shell_len_local_.push_back(this->shell_len(igsh));
+        for (auto igloc: it->second) {
+            gvec_shell_idx_local_[igloc] = num_gvec_shells_local_;
+        }
+        num_gvec_shells_local_++;
+    }
 }
 
 void Gvec::init_gvec_cart()
@@ -286,8 +314,8 @@ void Gvec::init_gvec_cart()
     for (int igloc = 0; igloc < count(); igloc++) {
         int ig   = offset() + igloc;
         auto G   = gvec_by_full_index(gvec_full_index_(ig));
-        auto gc  = lattice_vectors_ * vector3d<double>(G[0], G[1], G[2]);
-        auto gkc = lattice_vectors_ * (vector3d<double>(G[0], G[1], G[2]) + vk_);
+        auto gc  = dot(lattice_vectors_, vector3d<double>(G[0], G[1], G[2]));
+        auto gkc = dot(lattice_vectors_, (vector3d<double>(G[0], G[1], G[2]) + vk_));
         for (int x : {0, 1, 2}) {
             gvec_cart_(x, igloc)  = gc[x];
             gkvec_cart_(x, igloc) = gkc[x];
@@ -321,21 +349,21 @@ void Gvec::init(FFT3D_grid const& fft_grid)
         }
     }
     if (ig != num_gvec_) {
-        TERMINATE("wrong G-vector count");
+        throw std::runtime_error("wrong G-vector count");
     }
     for (int ig = 0; ig < num_gvec_; ig++) {
         auto gv = gvec(ig);
         if (index_by_gvec(gv) != ig) {
             std::stringstream s;
             s << "wrong G-vector index: ig=" << ig << " gv=" << gv << " index_by_gvec(gv)=" << index_by_gvec(gv);
-            TERMINATE(s);
+            throw std::runtime_error(s.str());
         }
     }
 
-    /* first G-vector must be (0, 0, 0); never reomove this check!!! */
+    /* first G-vector must be (0, 0, 0); never remove this check!!! */
     auto g0 = gvec_by_full_index(gvec_full_index_(0));
     if (g0[0] || g0[1] || g0[2]) {
-        TERMINATE("first G-vector is not zero");
+        throw std::runtime_error("first G-vector is not zero");
     }
 
     init_gvec_cart();
@@ -364,7 +392,7 @@ void Gvec::init(FFT3D_grid const& fft_grid)
                   << " G-vector index in new distribution : " << index_by_gvec(G) << std::endl
                   << " offset in G-vector index for this rank: " << offset() << std::endl
                   << " local number of G-vectors for this rank: " << count();
-                TERMINATE(s);
+                throw std::runtime_error(s.str());
             }
         }
     }
@@ -380,6 +408,8 @@ Gvec& Gvec::operator=(Gvec&& src__)
         reduce_gvec_       = src__.reduce_gvec_;
         bare_gvec_         = src__.bare_gvec_;
         num_gvec_          = src__.num_gvec_;
+        offset_            = src__.offset_;
+        count_             = src__.count_;
         gvec_full_index_   = std::move(src__.gvec_full_index_);
         gvec_shell_        = std::move(src__.gvec_shell_);
         num_gvec_shells_   = std::move(src__.num_gvec_shells_);
@@ -409,7 +439,7 @@ std::pair<int, bool> Gvec::index_g12_safe(vector3d<int> const& g1__, vector3d<in
           << "  G': " << g2__ << std::endl
           << "  G - G': " << v << std::endl
           << " idx: " << idx;
-        TERMINATE(s);
+        throw std::runtime_error(s.str());
     }
     return std::make_pair(idx, conj);
 }
@@ -452,57 +482,22 @@ int Gvec::index_by_gvec(vector3d<int> const& G__) const
     return ig;
 }
 
-void Gvec::pack(serializer& s__) const
-{
-    serialize(s__, vk_);
-    serialize(s__, Gmax_);
-    serialize(s__, lattice_vectors_);
-    serialize(s__, reduce_gvec_);
-    serialize(s__, bare_gvec_);
-    serialize(s__, num_gvec_);
-    serialize(s__, num_gvec_shells_);
-    serialize(s__, gvec_full_index_);
-    serialize(s__, gvec_shell_);
-    serialize(s__, gvec_shell_len_);
-    serialize(s__, gvec_index_by_xy_);
-    serialize(s__, z_columns_);
-    serialize(s__, gvec_distr_);
-    serialize(s__, zcol_distr_);
-    serialize(s__, gvec_base_mapping_);
-}
-
-void Gvec::unpack(serializer& s__, Gvec& gv__) const
-{
-    deserialize(s__, gv__.vk_);
-    deserialize(s__, gv__.Gmax_);
-    deserialize(s__, gv__.lattice_vectors_);
-    deserialize(s__, gv__.reduce_gvec_);
-    deserialize(s__, gv__.bare_gvec_);
-    deserialize(s__, gv__.num_gvec_);
-    deserialize(s__, gv__.num_gvec_shells_);
-    deserialize(s__, gv__.gvec_full_index_);
-    deserialize(s__, gv__.gvec_shell_);
-    deserialize(s__, gv__.gvec_shell_len_);
-    deserialize(s__, gv__.gvec_index_by_xy_);
-    deserialize(s__, gv__.z_columns_);
-    deserialize(s__, gv__.gvec_distr_);
-    deserialize(s__, gv__.zcol_distr_);
-    deserialize(s__, gv__.gvec_base_mapping_);
-}
-
-void Gvec::send_recv(Communicator const& comm__, int source__, int dest__, Gvec& gv__) const
+Gvec send_recv(Communicator const& comm__, Gvec const& gv_src__, int source__, int dest__)
 {
     serializer s;
 
     if (comm__.rank() == source__) {
-        this->pack(s);
+        ::sddk::serialize(s, gv_src__);
     }
 
     s.send_recv(comm__, source__, dest__);
 
+    Gvec gv(gv_src__.comm());
+
     if (comm__.rank() == dest__) {
-        this->unpack(s, gv__);
+        ::sddk::deserialize(s, gv);
     }
+    return gv;
 }
 
 void Gvec_partition::build_fft_distr()
@@ -572,11 +567,11 @@ Gvec_partition::Gvec_partition(Gvec const& gvec__, Communicator const& fft_comm_
 {
     if (fft_comm_.size() * comm_ortho_fft_.size() != gvec_.comm().size()) {
         std::stringstream s;
-        s << "wrong size of communicators" << std::endl
+        s << "[sddk::Gvec] wrong size of communicators" << std::endl
           << "  fft_comm_.size()       = " << fft_comm_.size() << std::endl
           << "  comm_ortho_fft_.size() = " << comm_ortho_fft_.size() << std::endl
           << "  gvec_.comm().size()    = " << gvec_.comm().size();
-        TERMINATE(s);
+        throw std::runtime_error(s.str());
     }
     rank_map_ = mdarray<int, 2>(fft_comm_.size(), comm_ortho_fft_.size());
     rank_map_.zero();
@@ -603,7 +598,7 @@ Gvec_partition::Gvec_partition(Gvec const& gvec__, Communicator const& fft_comm_
     int ig{0};
     for (int i = 0; i < comm_ortho_fft_.size(); i++) {
         for (int k = 0; k < gvec_.gvec_count(rank_map_(fft_comm_.rank(), i)); k++) {
-            idx_gvec_(ig) = gvec_.gvec_offset(rank_map_(fft_comm_.rank(), i)) + k;
+            idx_gvec_[ig] = gvec_.gvec_offset(rank_map_(fft_comm_.rank(), i)) + k;
             ig++;
         }
     }
@@ -612,22 +607,17 @@ Gvec_partition::Gvec_partition(Gvec const& gvec__, Communicator const& fft_comm_
     pile_gvec();
 }
 
-void Gvec_partition::gather_pw_fft(std::complex<double>* f_pw_local__, std::complex<double>* f_pw_fft__) const
+mdarray<int, 2> Gvec_partition::get_gvec() const
 {
-    int rank = gvec().comm().rank();
-    /* collect scattered PW coefficients */
-    comm_ortho_fft().allgather(f_pw_local__, gvec().gvec_count(rank), f_pw_fft__, gvec_fft_slab().counts.data(),
-                               gvec_fft_slab().offsets.data());
-}
-
-void Gvec_partition::gather_pw_global(std::complex<double>* f_pw_fft__, std::complex<double>* f_pw_global__) const
-{
-    for (int ig = 0; ig < gvec().count(); ig++) {
-        /* position inside fft buffer */
-        int ig1                             = gvec_fft_slab().offsets[comm_ortho_fft().rank()] + ig;
-        f_pw_global__[gvec().offset() + ig] = f_pw_fft__[ig1];
+    mdarray<int, 2> gv(3, gvec_count_fft());
+    for (int i = 0; i < gvec_count_fft(); i++) {
+        int ig = idx_gvec_[i];
+        auto G = gvec_.gvec(ig);
+        for (int x: {0, 1, 2}) {
+            gv(x, i) = G[x];
+        }
     }
-    gvec().comm().allgather(&f_pw_global__[0], gvec().offset(), gvec().count());
+    return gv;
 }
 
 Gvec_shells::Gvec_shells(Gvec const& gvec__)
@@ -636,23 +626,23 @@ Gvec_shells::Gvec_shells(Gvec const& gvec__)
 {
     PROFILE("sddk::Gvec_shells");
 
-    a2a_send = block_data_descriptor(comm_.size());
-    a2a_recv = block_data_descriptor(comm_.size());
+    a2a_send_ = block_data_descriptor(comm_.size());
+    a2a_recv_ = block_data_descriptor(comm_.size());
 
     /* split G-vector shells between ranks in cyclic order */
-    spl_num_gsh = splindex<splindex_t::block_cyclic>(gvec_.num_shells(), comm_.size(), comm_.rank(), 1);
+    spl_num_gsh_ = splindex<splindex_t::block_cyclic>(gvec_.num_shells(), comm_.size(), comm_.rank(), 1);
 
     /* each rank sends a fraction of its local G-vectors to other ranks */
     /* count this fraction */
     for (int igloc = 0; igloc < gvec_.count(); igloc++) {
         int ig   = gvec_.offset() + igloc;
         int igsh = gvec_.shell(ig);
-        a2a_send.counts[spl_num_gsh.local_rank(igsh)]++;
+        a2a_send_.counts[spl_num_gsh_.local_rank(igsh)]++;
     }
-    a2a_send.calc_offsets();
+    a2a_send_.calc_offsets();
     /* sanity check: total number of elements to send is equal to the local number of G-vector */
-    if (a2a_send.size() != gvec_.count()) {
-        TERMINATE("wrong number of G-vectors");
+    if (a2a_send_.size() != gvec_.count()) {
+        throw std::runtime_error("wrong number of G-vectors");
     }
     /* count the number of elements to receive */
     for (int r = 0; r < comm_.size(); r++) {
@@ -661,39 +651,93 @@ Gvec_shells::Gvec_shells(Gvec const& gvec__)
             int ig = gvec_.gvec_offset(r) + igloc;
             /* index of the G-vector shell */
             int igsh = gvec_.shell(ig);
-            if (spl_num_gsh.local_rank(igsh) == comm_.rank()) {
-                a2a_recv.counts[r]++;
+            if (spl_num_gsh_.local_rank(igsh) == comm_.rank()) {
+                a2a_recv_.counts[r]++;
             }
         }
     }
-    a2a_recv.calc_offsets();
+    a2a_recv_.calc_offsets();
     /* sanity check: sum of local sizes in the remapped order is equal to the total number of G-vectors */
     int ng = gvec_count_remapped();
     comm_.allreduce(&ng, 1);
     if (ng != gvec_.num_gvec()) {
-        TERMINATE("wrong number of G-vectors");
+        throw std::runtime_error("wrong number of G-vectors");
     }
 
     /* local set of G-vectors in the remapped order */
-    gvec_remapped_       = mdarray<int, 2>(3, gvec_count_remapped());
-    gvec_shell_remapped_ = mdarray<int, 1>(gvec_count_remapped());
+    gvec_remapped_       = sddk::mdarray<int, 2>(3, gvec_count_remapped());
+    gvec_shell_remapped_ = sddk::mdarray<int, 1>(gvec_count_remapped());
     std::vector<int> counts(comm_.size(), 0);
     for (int r = 0; r < comm_.size(); r++) {
         for (int igloc = 0; igloc < gvec_.gvec_count(r); igloc++) {
             int ig   = gvec_.gvec_offset(r) + igloc;
             int igsh = gvec_.shell(ig);
             auto G   = gvec_.gvec(ig);
-            if (spl_num_gsh.local_rank(igsh) == comm_.rank()) {
+            if (spl_num_gsh_.local_rank(igsh) == comm_.rank()) {
                 for (int x : {0, 1, 2}) {
-                    gvec_remapped_(x, a2a_recv.offsets[r] + counts[r]) = G[x];
+                    gvec_remapped_(x, a2a_recv_.offsets[r] + counts[r]) = G[x];
                 }
-                gvec_shell_remapped_(a2a_recv.offsets[r] + counts[r]) = igsh;
+                gvec_shell_remapped_(a2a_recv_.offsets[r] + counts[r]) = igsh;
                 counts[r]++;
             }
         }
     }
     for (int ig = 0; ig < gvec_count_remapped(); ig++) {
-        idx_gvec[gvec_remapped(ig)] = ig;
+        idx_gvec_[gvec_remapped(ig)] = ig;
+    }
+    /* sanity check */
+    for (int igloc = 0; igloc < this->gvec_count_remapped(); igloc++) {
+        auto G = this->gvec_remapped(igloc);
+        if (this->index_by_gvec(G) != igloc) {
+            throw std::runtime_error("Wrong remapped index of G-vector");
+        }
+        int igsh = this->gvec_shell_remapped(igloc);
+        if (igsh != this->gvec().shell(G)) {
+            throw std::runtime_error("Wrong remapped shell of G-vector");
+        }
     }
 }
+
+void serialize(serializer& s__, Gvec const& gv__)
+{
+    serialize(s__, gv__.vk_);
+    serialize(s__, gv__.Gmax_);
+    serialize(s__, gv__.lattice_vectors_);
+    serialize(s__, gv__.reduce_gvec_);
+    serialize(s__, gv__.bare_gvec_);
+    serialize(s__, gv__.num_gvec_);
+    serialize(s__, gv__.num_gvec_shells_);
+    serialize(s__, gv__.gvec_full_index_);
+    serialize(s__, gv__.gvec_shell_);
+    serialize(s__, gv__.gvec_shell_len_);
+    serialize(s__, gv__.gvec_index_by_xy_);
+    serialize(s__, gv__.z_columns_);
+    serialize(s__, gv__.gvec_distr_);
+    serialize(s__, gv__.zcol_distr_);
+    serialize(s__, gv__.gvec_base_mapping_);
+    serialize(s__, gv__.offset_);
+    serialize(s__, gv__.count_);
+}
+
+void deserialize(serializer& s__, Gvec& gv__)
+{
+    deserialize(s__, gv__.vk_);
+    deserialize(s__, gv__.Gmax_);
+    deserialize(s__, gv__.lattice_vectors_);
+    deserialize(s__, gv__.reduce_gvec_);
+    deserialize(s__, gv__.bare_gvec_);
+    deserialize(s__, gv__.num_gvec_);
+    deserialize(s__, gv__.num_gvec_shells_);
+    deserialize(s__, gv__.gvec_full_index_);
+    deserialize(s__, gv__.gvec_shell_);
+    deserialize(s__, gv__.gvec_shell_len_);
+    deserialize(s__, gv__.gvec_index_by_xy_);
+    deserialize(s__, gv__.z_columns_);
+    deserialize(s__, gv__.gvec_distr_);
+    deserialize(s__, gv__.zcol_distr_);
+    deserialize(s__, gv__.gvec_base_mapping_);
+    deserialize(s__, gv__.offset_);
+    deserialize(s__, gv__.count_);
+}
+
 } // namespace sddk
