@@ -28,16 +28,14 @@
 #include <numeric>
 #include <map>
 #include <iostream>
+#include <type_traits>
 #include <assert.h>
 #include "memory.hpp"
 #include "fft3d_grid.hpp"
 #include "geometry3d.hpp"
 #include "serializer.hpp"
 #include "splindex.hpp"
-#include "utils/utils.hpp"
 #include "utils/profiler.hpp"
-//#include "Unit_cell/unit_cell_symmetry.hpp"
-#include "Symmetry/find_lat_sym.hpp"
 
 using namespace geometry3d;
 
@@ -107,6 +105,12 @@ inline void deserialize(serializer& s__, std::vector<z_column_descriptor>& zcol_
     }
 }
 
+/* forward declarations */
+class Gvec;
+void serialize(serializer& s__, Gvec const& gv__);
+void deserialize(serializer& s__, Gvec& gv__);
+Gvec send_recv(Communicator const& comm__, Gvec const& gv_src__, int source__, int dest__);
+
 /// A set of G-vectors for FFTs and G+k basis functions.
 /** Current implemntation supports up to 2^12 (4096) z-dimension of the FFT grid and 2^20 (1048576) number of
  *  z-columns. The order of z-sticks and G-vectors is not fixed and depends on the number of MPI ranks used
@@ -151,9 +155,22 @@ class Gvec
     /// Number of G-vector shells (groups of G-vectors with the same length).
     int num_gvec_shells_;
 
-    mdarray<double, 1> gvec_shell_len_;
+    /// Radii (or lengths) of G-vector shells in a.u.^-1.
+    sddk::mdarray<double, 1> gvec_shell_len_;
 
-    mdarray<int, 3> gvec_index_by_xy_;
+    /// Local number of G-vector shells for the local number of G-vectors.
+    /** G-vectors are distributed by sticks, not by G-shells. This means that each rank stores local fraction of
+        G-vectors with a non-consecutive G-shell index and not all G-shells are present at a given rank. This
+        variable stores the number of G-shells which this rank holds. */
+    int num_gvec_shells_local_;
+
+    /// Radii of G-vector shells in the local index counting [0, num_gvec_shells_local)
+    std::vector<double> gvec_shell_len_local_;
+
+    /// Mapping between local index of G-vector and local  G-shell index.
+    std::vector<int> gvec_shell_idx_local_;
+
+    sddk::mdarray<int, 3> gvec_index_by_xy_;
 
     /// Global list of non-zero z-columns.
     std::vector<z_column_descriptor> z_columns_;
@@ -188,15 +205,15 @@ class Gvec
     /// Cartesian coordinaes for a local set of G+k-vectors.
     mdarray<double, 2> gkvec_cart_;
 
-    /* copy constructor is forbidden */
-    Gvec(Gvec const& src__) = delete;
-
-    /* copy assigment operator is forbidden */
-    Gvec& operator=(Gvec const& src__) = delete;
-
     /// Return corresponding G-vector for an index in the range [0, num_gvec).
     vector3d<int> gvec_by_full_index(uint32_t idx__) const;
 
+    /// Offset in the global index for the local part of G-vectors.
+    int offset_{-1};
+
+    /// Local number of G-vectors.
+    int count_{-1};
+
     /// Find z-columns of G-vectors inside a sphere with Gmax radius.
     /** This function also computes the total number of G-vectors. */
     void find_z_columns(double Gmax__, FFT3D_grid const& fft_box__);
@@ -206,7 +223,7 @@ class Gvec
 
     /// Find a list of G-vector shells.
     /** G-vectors belonging to the same shell have the same length and transform to each other
-        under a lattice symmetry operation
+        under a lattice symmetry operation.
      */
     void find_gvec_shells();
 
@@ -216,6 +233,16 @@ class Gvec
     /// Initialize everything.
     void init(FFT3D_grid const& fft_grid);
 
+    friend void sddk::serialize(serializer& s__, Gvec const& gv__);
+
+    friend void sddk::deserialize(serializer& s__, Gvec& gv__);
+
+    /* copy constructor is forbidden */
+    Gvec(Gvec const& src__) = delete;
+
+    /* copy assignment operator is forbidden */
+    Gvec& operator=(Gvec const& src__) = delete;
+
   public:
     /// Constructor for G+k vectors.
     /** \param [in] vk          K-point vector of G+k
@@ -294,21 +321,22 @@ class Gvec
     {
     }
 
-    /// Move assigment operator.
+    /// Move assignment operator.
     Gvec& operator=(Gvec&& src__);
 
+    /// Move constructor.
     Gvec(Gvec&& src__)
         : comm_(src__.comm_)
     {
         *this = std::move(src__);
     }
 
-    inline vector3d<double> const& vk() const
+    inline auto const& vk() const
     {
         return vk_;
     }
 
-    Communicator const& comm() const
+    inline Communicator const& comm() const
     {
         return comm_;
     }
@@ -316,7 +344,7 @@ class Gvec
     /// Set the new reciprocal lattice vectors.
     /** For the varibale-cell relaxation runs we need an option to preserve the number of G- and G+k vectors.
      *  Here we can set the new lattice vectors and update the relevant members of the Gvec class. */
-    inline matrix3d<double> const& lattice_vectors(matrix3d<double> lattice_vectors__)
+    inline auto const& lattice_vectors(matrix3d<double> lattice_vectors__)
     {
         lattice_vectors_ = lattice_vectors__;
         init_gvec_cart();
@@ -368,7 +396,7 @@ class Gvec
     /** The \em count and \em offset are borrowed from the MPI terminology for data distribution. */
     inline int count() const
     {
-        return gvec_count(comm().rank());
+        return count_;
     }
 
     /// Offset (in the global index) of G-vectors for a fine-grained distribution.
@@ -382,7 +410,7 @@ class Gvec
     /** The \em count and \em offset are borrowed from the MPI terminology for data distribution. */
     inline int offset() const
     {
-        return gvec_offset(comm().rank());
+        return offset_;
     }
 
     /// Local starting index of G-vectors if G=0 is not counted.
@@ -398,13 +426,13 @@ class Gvec
     }
 
     /// Return G vector in fractional coordinates.
-    inline vector3d<int> gvec(int ig__) const
+    inline auto gvec(int ig__) const
     {
         return gvec_by_full_index(gvec_full_index_(ig__));
     }
 
     /// Return G+k vector in fractional coordinates.
-    inline vector3d<double> gkvec(int ig__) const
+    inline auto gkvec(int ig__) const
     {
         auto G = gvec_by_full_index(gvec_full_index_(ig__));
         return (vector3d<double>(G[0], G[1], G[2]) + vk_);
@@ -412,44 +440,66 @@ class Gvec
 
     /// Return G vector in Cartesian coordinates.
     template <index_domain_t idx_t>
-    inline vector3d<double> gvec_cart(int ig__) const
+    inline std::enable_if_t<idx_t == index_domain_t::local, vector3d<double>>
+    gvec_cart(int ig__) const
     {
-        switch (idx_t) {
-            case index_domain_t::local: {
-                return vector3d<double>(gvec_cart_(0, ig__), gvec_cart_(1, ig__), gvec_cart_(2, ig__));
-            }
-            case index_domain_t::global: {
-                auto G = gvec_by_full_index(gvec_full_index_(ig__));
-                return lattice_vectors_ * vector3d<double>(G[0], G[1], G[2]);
-            }
-        }
+        return vector3d<double>(gvec_cart_(0, ig__), gvec_cart_(1, ig__), gvec_cart_(2, ig__));
+    }
+
+    /// Return G vector in Cartesian coordinates.
+    template <index_domain_t idx_t>
+    inline std::enable_if_t<idx_t == index_domain_t::global, vector3d<double>>
+    gvec_cart(int ig__) const
+    {
+        auto G = gvec_by_full_index(gvec_full_index_(ig__));
+        return dot(lattice_vectors_, vector3d<double>(G[0], G[1], G[2]));
     }
 
     /// Return G+k vector in Cartesian coordinates.
     template <index_domain_t idx_t>
-    inline vector3d<double> gkvec_cart(int ig__) const
+    inline std::enable_if_t<idx_t == index_domain_t::local, vector3d<double>>
+    gkvec_cart(int ig__) const
     {
-        switch (idx_t) {
-            case index_domain_t::local: {
-                return vector3d<double>(gkvec_cart_(0, ig__), gkvec_cart_(1, ig__), gkvec_cart_(2, ig__));
-            }
-            case index_domain_t::global: {
-                auto G = gvec_by_full_index(gvec_full_index_(ig__));
-                return lattice_vectors_ * (vector3d<double>(G[0], G[1], G[2]) + vk_);
-            }
-        }
+        return vector3d<double>(gkvec_cart_(0, ig__), gkvec_cart_(1, ig__), gkvec_cart_(2, ig__));
     }
 
+    /// Return G+k vector in Cartesian coordinates.
+    template <index_domain_t idx_t>
+    inline std::enable_if_t<idx_t == index_domain_t::global, vector3d<double>>
+    gkvec_cart(int ig__) const
+    {
+        auto G = gvec_by_full_index(gvec_full_index_(ig__));
+        return dot(lattice_vectors_, vector3d<double>(G[0], G[1], G[2]) + vk_);
+    }
+
+    /// Return index of the G-vector shell by the G-vector index.
     inline int shell(int ig__) const
     {
         return gvec_shell_(ig__);
     }
 
+    inline int shell(vector3d<int> const& G__) const
+    {
+        return this->shell(index_by_gvec(G__));
+    }
+
+    /// Return length of the G-vector shell.
     inline double shell_len(int igs__) const
     {
         return gvec_shell_len_(igs__);
     }
 
+    /// Get lengths of all G-vector shells.
+    std::vector<double> shells_len() const
+    {
+        std::vector<double> q(this->num_shells());
+        for (int i = 0; i < this->num_shells(); i++) {
+            q[i] = this->shell_len(i);
+        }
+        return q;
+    }
+
+    /// Return length of the G-vector.
     inline double gvec_len(int ig__) const
     {
         return gvec_shell_len_(gvec_shell_(ig__));
@@ -466,11 +516,11 @@ class Gvec
 
     std::pair<int, bool> index_g12_safe(vector3d<int> const& g1__, vector3d<int> const& g2__) const;
 
-    inline int index_g12_safe(int ig1__, int ig2__) const
-    {
-        STOP();
-        return 0;
-    }
+    //inline int index_g12_safe(int ig1__, int ig2__) const
+    //{
+    //    STOP();
+    //    return 0;
+    //}
 
     /// Return a global G-vector index in the range [0, num_gvec) by the G-vector.
     /** The information about a G-vector index is encoded by two numbers: a starting index for the
@@ -507,68 +557,22 @@ class Gvec
         return gvec_base_mapping_(igloc_base__);
     }
 
-    friend void serialize(serializer& s__, Gvec& gv__);
-    friend void deserialize(serializer& s__, Gvec& gv__);
+    inline int num_gvec_shells_local() const
+    {
+        return num_gvec_shells_local_;
+    }
 
-    /// Serialize to a string of bytes.
-    void pack(serializer& s__) const;
+    inline double gvec_shell_len_local(int idx__) const
+    {
+        return gvec_shell_len_local_[idx__];
+    }
 
-    /// Deserialize from a string of bytes.
-    void unpack(serializer& s__, Gvec& gv__) const;
-
-    void send_recv(Communicator const& comm__, int source__, int dest__, Gvec& gv__) const;
-
-    //friend std::unique_ptr<Gvec> send_recv(Gvec const& gv__, Communicator const& comm__, int source__, int dest__);
+    inline int gvec_shell_idx_local(int igloc__) const
+    {
+        return gvec_shell_idx_local_[igloc__];
+    }
 };
 
-//inline std::unique_ptr<Gvec> send_recv(Gvec const& gv__, Communicator const& comm__, int source__, int dest__)
-//{
-//    std::unique_ptr<Gvec> gvout(new Gvec(gv__.comm()));
-//
-//    serializer s;
-//
-//    if (comm__.rank() == source__) {
-//        std::cout << "address of gv: " << &gv__ << "\n";
-//        serialize(s, gv__.vk_);
-//        serialize(s, gv__.Gmax_);
-//        serialize(s, gv__.lattice_vectors_);
-//        serialize(s, gv__.reduce_gvec_);
-//        serialize(s, gv__.bare_gvec_);
-//        serialize(s, gv__.num_gvec_);
-//        serialize(s, gv__.num_gvec_shells_);
-//        serialize(s, gv__.gvec_full_index_);
-//        serialize(s, gv__.gvec_shell_);
-//        serialize(s, gv__.gvec_shell_len_);
-//        serialize(s, gv__.gvec_index_by_xy_);
-//        serialize(s, gv__.z_columns_);
-//        serialize(s, gv__.gvec_distr_);
-//        serialize(s, gv__.zcol_distr_);
-//        serialize(s, gv__.gvec_base_mapping_);
-//    }
-//
-//    s.send_recv(comm__, source__, dest__);
-//
-//    if (comm__.rank() == dest__) {
-//        deserialize(s, gvout->vk_);
-//        deserialize(s, gvout->Gmax_);
-//        deserialize(s, gvout->lattice_vectors_);
-//        deserialize(s, gvout->reduce_gvec_);
-//        deserialize(s, gvout->bare_gvec_);
-//        deserialize(s, gvout->num_gvec_);
-//        deserialize(s, gvout->num_gvec_shells_);
-//        deserialize(s, gvout->gvec_full_index_);
-//        deserialize(s, gvout->gvec_shell_);
-//        deserialize(s, gvout->gvec_shell_len_);
-//        deserialize(s, gvout->gvec_index_by_xy_);
-//        deserialize(s, gvout->z_columns_);
-//        deserialize(s, gvout->gvec_distr_);
-//        deserialize(s, gvout->zcol_distr_);
-//        deserialize(s, gvout->gvec_base_mapping_);
-//    }
-//
-//    return std::move(gvout);
-//}
-
 /// Stores information about G-vector partitioning between MPI ranks for the FFT transformation.
 /** FFT driver works with a small communicator. G-vectors are distributed over the entire communicator which is
     larger than the FFT communicator. In order to transform the functions, G-vectors must be redistributed to the
@@ -601,7 +605,7 @@ class Gvec_partition
     /// Mapping of MPI ranks used to split G-vectors to a 2D grid.
     mdarray<int, 2> rank_map_;
 
-    /// Global index of z-column in new (fat-slab) distrubution.
+    /// Global index of z-column in new (fat-slab) distribution.
     /** This is a mapping between new and original ordering of z-columns. */
     mdarray<int, 1> idx_zcol_;
 
@@ -635,7 +639,7 @@ class Gvec_partition
         return gvec_distr_fft_.counts[rank__];
     }
 
-    /// Local number of G-vectors for FFT-friendly distibution.
+    /// Local number of G-vectors for FFT-friendly distribution.
     inline int gvec_count_fft() const
     {
         return gvec_count_fft(fft_comm().rank());
@@ -687,9 +691,28 @@ class Gvec_partition
         return gvec_;
     }
 
-    void gather_pw_fft(std::complex<double>* f_pw_local__, std::complex<double>* f_pw_fft__) const;
+    mdarray<int, 2> get_gvec() const;
 
-    void gather_pw_global(std::complex<double>* f_pw_fft__, std::complex<double>* f_pw_global__) const;
+    template <typename T>
+    void gather_pw_fft(std::complex<T>* f_pw_local__, std::complex<T>* f_pw_fft__) const
+    {
+        int rank = gvec().comm().rank();
+        /* collect scattered PW coefficients */
+        comm_ortho_fft().allgather(f_pw_local__, gvec().gvec_count(rank), f_pw_fft__, gvec_fft_slab().counts.data(),
+                                   gvec_fft_slab().offsets.data());
+
+    }
+
+    template <typename T>
+    void gather_pw_global(std::complex<T>* f_pw_fft__, std::complex<T>* f_pw_global__) const
+    {
+        for (int ig = 0; ig < gvec().count(); ig++) {
+            /* position inside fft buffer */
+            int ig1                             = gvec_fft_slab().offsets[comm_ortho_fft().rank()] + ig;
+            f_pw_global__[gvec().offset() + ig] = f_pw_fft__[ig1];
+        }
+        gvec().comm().allgather(&f_pw_global__[0], gvec().count(), gvec().offset());
+    }
 };
 
 /// Helper class to manage G-vector shells and redistribute G-vectors for symmetrization.
@@ -701,19 +724,19 @@ class Gvec_shells
 {
   private:
     /// Sending counts and offsets.
-    block_data_descriptor a2a_send;
+    block_data_descriptor a2a_send_;
 
     /// Receiving counts and offsets.
-    block_data_descriptor a2a_recv;
+    block_data_descriptor a2a_recv_;
 
     /// Split global index of G-shells between MPI ranks.
-    splindex<splindex_t::block_cyclic> spl_num_gsh;
+    splindex<splindex_t::block_cyclic> spl_num_gsh_;
 
     /// List of G-vectors in the remapped storage.
-    mdarray<int, 2> gvec_remapped_;
+    sddk::mdarray<int, 2> gvec_remapped_;
 
     /// Mapping between index of local G-vector and global index of G-vector shell.
-    mdarray<int, 1> gvec_shell_remapped_;
+    sddk::mdarray<int, 1> gvec_shell_remapped_;
 
     /// Alias for the G-vector communicator.
     Communicator const& comm_;
@@ -721,7 +744,7 @@ class Gvec_shells
     Gvec const& gvec_;
 
     /// A mapping between G-vector and it's local index in the new distribution.
-    std::map<vector3d<int>, int> idx_gvec;
+    std::map<vector3d<int>, int> idx_gvec_;
 
   public:
 
@@ -731,18 +754,23 @@ class Gvec_shells
     {
         pstdout pout(gvec_.comm());
         pout.printf("rank: %i\n", gvec_.comm().rank());
+        pout.printf("-- list of G-vectors in the remapped distribution --\n");
         for (int igloc = 0; igloc < gvec_count_remapped(); igloc++) {
             auto G = gvec_remapped(igloc);
 
             int igsh = gvec_shell_remapped(igloc);
             pout.printf("igloc=%i igsh=%i G=%i %i %i\n", igloc, igsh, G[0], G[1], G[2]);
         }
+        pout.printf("-- reverse list --\n");
+        for (auto const& e: idx_gvec_) {
+            pout.printf("G=%i %i %i, igloc=%i\n", e.first[0], e.first[1], e.first[2], e.second);
+        }
     }
 
     /// Local number of G-vectors in the remapped distribution with complete shells on each rank.
     int gvec_count_remapped() const
     {
-        return a2a_recv.size();
+        return a2a_recv_.size();
     }
 
     /// G-vector by local index (in the remapped set).
@@ -754,8 +782,8 @@ class Gvec_shells
     /// Return local index of the G-vector in the remapped set.
     int index_by_gvec(vector3d<int> G__) const
     {
-        if (idx_gvec.count(G__)) {
-            return idx_gvec.at(G__);
+        if (idx_gvec_.count(G__)) {
+            return idx_gvec_.at(G__);
         } else {
             return -1;
         }
@@ -775,17 +803,17 @@ class Gvec_shells
         std::vector<T> send_buf(gvec_.count());
         std::vector<int> counts(comm_.size(), 0);
         for (int igloc = 0; igloc < gvec_.count(); igloc++) {
-            int ig                                    = gvec_.offset() + igloc;
-            int igsh                                  = gvec_.shell(ig);
-            int r                                     = spl_num_gsh.local_rank(igsh);
-            send_buf[a2a_send.offsets[r] + counts[r]] = data__[igloc];
+            int ig                                     = gvec_.offset() + igloc;
+            int igsh                                   = gvec_.shell(ig);
+            int r                                      = spl_num_gsh_.local_rank(igsh);
+            send_buf[a2a_send_.offsets[r] + counts[r]] = data__[igloc];
             counts[r]++;
         }
 
         std::vector<T> recv_buf(gvec_count_remapped());
 
-        comm_.alltoall(send_buf.data(), a2a_send.counts.data(), a2a_send.offsets.data(), recv_buf.data(),
-                       a2a_recv.counts.data(), a2a_recv.offsets.data());
+        comm_.alltoall(send_buf.data(), a2a_send_.counts.data(), a2a_send_.offsets.data(), recv_buf.data(),
+                       a2a_recv_.counts.data(), a2a_recv_.offsets.data());
 
         return recv_buf;
     }
@@ -797,15 +825,15 @@ class Gvec_shells
 
         std::vector<T> recv_buf(gvec_.count());
 
-        comm_.alltoall(buf__.data(), a2a_recv.counts.data(), a2a_recv.offsets.data(), recv_buf.data(),
-                       a2a_send.counts.data(), a2a_send.offsets.data());
+        comm_.alltoall(buf__.data(), a2a_recv_.counts.data(), a2a_recv_.offsets.data(), recv_buf.data(),
+                       a2a_send_.counts.data(), a2a_send_.offsets.data());
 
         std::vector<int> counts(comm_.size(), 0);
         for (int igloc = 0; igloc < gvec_.count(); igloc++) {
             int ig        = gvec_.offset() + igloc;
             int igsh      = gvec_.shell(ig);
-            int r         = spl_num_gsh.local_rank(igsh);
-            data__[igloc] = recv_buf[a2a_send.offsets[r] + counts[r]];
+            int r         = spl_num_gsh_.local_rank(igsh);
+            data__[igloc] = recv_buf[a2a_send_.offsets[r] + counts[r]];
             counts[r]++;
         }
     }
