@@ -35,16 +35,18 @@
 #include <array>
 #include <complex>
 #include <cassert>
-#include "GPU/acc.hpp"
+#include "gpu/acc.hpp"
 
 namespace sddk {
 
+/// Check is the type is a complex number; by default it is not.
 template <typename T>
 struct is_complex
 {
     constexpr static bool value{false};
 };
 
+/// Check is the type is a complex number: for std::complex<T> it is true.
 template<typename T>
 struct is_complex<std::complex<T>>
 {
@@ -84,22 +86,20 @@ inline bool is_device_memory(memory_t mem__)
 inline memory_t get_memory_t(std::string name__)
 {
     std::transform(name__.begin(), name__.end(), name__.begin(), ::tolower);
-
-    std::map<std::string, memory_t> const map_to_type = {
-        {"none",        memory_t::none},
-        {"host",        memory_t::host},
+    std::map<std::string, memory_t> const m = {
+        {"none", memory_t::none},
+        {"host", memory_t::host},
         {"host_pinned", memory_t::host_pinned},
-        {"managed",     memory_t::managed},
-        {"device",      memory_t::device}
+        {"managed", memory_t::managed},
+        {"device", memory_t::device}
     };
 
-    if (map_to_type.count(name__) == 0) {
+    if (m.count(name__) == 0) {
         std::stringstream s;
-        s << "wrong label of memory type: " << name__;
+        s << "get_memory_t(): wrong label of the memory_t enumerator: " << name__;
         throw std::runtime_error(s.str());
-    }
-
-    return map_to_type.at(name__);
+     }
+     return m.at(name__);
 }
 
 /// Type of the main processing unit.
@@ -131,17 +131,21 @@ inline device_t get_device_t(memory_t mem__)
     return device_t::CPU; // make compiler happy
 }
 
+/// Get device type from the string.
 inline device_t get_device_t(std::string name__)
 {
     std::transform(name__.begin(), name__.end(), name__.begin(), ::tolower);
-    if (name__ == "cpu") {
-        return device_t::CPU;
-    } else if (name__ == "gpu") {
-        return device_t::GPU;
-    } else {
-        throw std::runtime_error("get_device_t(): wrong processing unit");
-    }
-    return device_t::CPU; // make compiler happy
+    std::map<std::string, device_t> const m = {
+        {"cpu", device_t::CPU},
+        {"gpu", device_t::GPU}
+    };
+
+    if (m.count(name__) == 0) {
+        std::stringstream s;
+        s << "get_device_t(): wrong label of the device_t enumerator: " << name__;
+        throw std::runtime_error(s.str());
+     }
+     return m.at(name__);
 }
 
 /// Allocate n elements in a specified memory.
@@ -158,14 +162,14 @@ inline T* allocate(size_t n__, memory_t M__)
             return static_cast<T*>(std::malloc(n__ * sizeof(T)));
         }
         case memory_t::host_pinned: {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
             return acc::allocate_host<T>(n__);
 #else
             return nullptr;
 #endif
         }
         case memory_t::device: {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
             return acc::allocate<T>(n__);
 #else
             return nullptr;
@@ -189,13 +193,13 @@ inline void deallocate(void* ptr__, memory_t M__)
             break;
         }
         case memory_t::host_pinned: {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
             acc::deallocate_host(ptr__);
 #endif
             break;
         }
         case memory_t::device: {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
             acc::deallocate(ptr__);
 #endif
             break;
@@ -207,6 +211,19 @@ inline void deallocate(void* ptr__, memory_t M__)
 }
 
 /// Copy between different memory types.
+template <typename T, typename F>
+inline void copy(memory_t from_mem__, T const* from_ptr__, memory_t to_mem__, F* to_ptr__, size_t n__)
+{
+    if (is_host_memory(to_mem__) && is_host_memory(from_mem__)) {
+        std::copy(from_ptr__, from_ptr__ + n__, to_ptr__);
+        return;
+    }
+#if defined(SIRIUS_GPU)
+    throw std::runtime_error("Copy mixed precision type not supported in device memory");
+    return;
+#endif
+}
+
 template <typename T>
 inline void copy(memory_t from_mem__, T const* from_ptr__, memory_t to_mem__, T* to_ptr__, size_t n__)
 {
@@ -214,7 +231,7 @@ inline void copy(memory_t from_mem__, T const* from_ptr__, memory_t to_mem__, T*
         std::memcpy(to_ptr__, from_ptr__, n__ * sizeof(T));
         return;
     }
-#if defined(__GPU)
+#if defined(SIRIUS_GPU)
     if (is_device_memory(to_mem__) && is_device_memory(from_mem__)) {
         acc::copy(to_ptr__, from_ptr__, n__);
         return;
@@ -453,7 +470,7 @@ struct memory_subblock_descriptor
 };
 
 //// Memory pool.
-/** This class stores list of allocated memory blocks. Each of the blocks can be devided into subblocks. When subblock
+/** This class stores list of allocated memory blocks. Each of the blocks can be divided into subblocks. When subblock
  *  is deallocated it is merged with previous or next free subblock in the memory block. If this was the last subblock
  *  in the block of memory, the (now) free block of memory is merged with the neighbours (if any are available).
  */
@@ -482,8 +499,9 @@ class memory_pool
     template <typename T>
     T* allocate(size_t num_elements__)
     {
+#if defined(SIRIUS_USE_MEMORY_POOL)
         /* memory block descriptor returns an unaligned memory; here we compute the the aligment value */
-        size_t align_size = std::max(size_t(64), alignof(T));
+        size_t align_size = std::max(size_t(SIRIUS_GPU_MEMORY_ALIGMENT), alignof(T));
         /* size of the memory block in bytes */
         size_t size = num_elements__ * sizeof(T) + align_size;
 
@@ -541,54 +559,36 @@ class memory_pool
         /* add to the hash table */
         map_ptr_[aligned_ptr] = msb;
         return reinterpret_cast<T*>(aligned_ptr);
+#else
+        return sddk::allocate<T>(num_elements__, M_);
+#endif
     }
 
     /// Delete a pointer and add its memory back to the pool.
     void free(void* ptr__)
     {
+#if defined(SIRIUS_USE_MEMORY_POOL)
         auto ptr = reinterpret_cast<uint8_t*>(ptr__);
         /* get a descriptor of this pointer */
         auto& msb = map_ptr_.at(ptr);
         /* free the sub-block */
         msb.it_->free_subblock(msb.unaligned_ptr_, msb.size_);
-
-        //auto merge_blocks = [&](std::list<memory_block_descriptor>::iterator it0,
-        //                        std::list<memory_block_descriptor>::iterator it)
-        //{
-        //    if (it0->is_empty()) {
-        //        size_t size = it->size_ + it0->size_;
-        //        it->buffer_ = nullptr;
-        //        it0->buffer_ = nullptr;
-        //        (*it) = memory_block_descriptor(size, M_);
-        //        memory_blocks_.erase(it0);
-        //    }
-        //};
-
-        ///* merge memory blocks; this is not strictly necessary but can lead to a better performance */
-        //auto it = msb.it_;
-        //if (it->is_empty()) {
-        //    /* try the previous block */
-        //    if (it != memory_blocks_.begin()) {
-        //        auto it0 = it;
-        //        it0--;
-        //        merge_blocks(it0, it);
-        //    }
-        //    /* try the next block */
-        //    auto it0 = it;
-        //    it0++;
-        //    if (it0 != memory_blocks_.end()) {
-        //        merge_blocks(it0, it);
-        //    }
-        //}
         /* remove this pointer from the hash table */
         map_ptr_.erase(ptr);
+#else
+        sddk::deallocate(ptr__, M_);
+#endif
     }
 
     /// Return a unique pointer to the allocated memory.
     template <typename T>
     std::unique_ptr<T, memory_t_deleter_base> get_unique_ptr(size_t n__)
     {
-        return std::unique_ptr<T, memory_t_deleter_base>(allocate<T>(n__), memory_pool_deleter(this));
+#if defined(SIRIUS_USE_MEMORY_POOL)
+        return std::unique_ptr<T, memory_t_deleter_base>(this->allocate<T>(n__), memory_pool_deleter(this));
+#else
+        return sddk::get_unique_ptr<T>(n__, M_);
+#endif
     }
 
     /// Free all the allocated blocks.
@@ -661,7 +661,7 @@ void memory_pool_deleter::memory_pool_deleter_impl::free(void* ptr__)
     mp_->free(ptr__);
 }
 
-//#ifdef __GPU
+//#ifdef SIRIUS_GPU
 //extern "C" void add_checksum_gpu(cuDoubleComplex* wf__,
 //                                 int num_rows_loc__,
 //                                 int nwf__,
@@ -671,19 +671,20 @@ void memory_pool_deleter::memory_pool_deleter_impl::free(void* ptr__)
 #ifdef NDEBUG
 #define mdarray_assert(condition__)
 #else
-#define mdarray_assert(condition__)                             \
-{                                                               \
-    if (!(condition__)) {                                       \
-        printf("Assertion (%s) failed ", #condition__);         \
-        printf("at line %i of file %s\n", __LINE__, __FILE__);  \
-        printf("array label: %s\n", label_.c_str());            \
-        for (int i = 0; i < N; i++) {                           \
-            printf("dim[%i].size = %li\n", i, dims_[i].size()); \
-        }                                                       \
-        raise(SIGTERM);                                         \
-        exit(-13);                                              \
-    }                                                           \
-}
+#define mdarray_assert(condition__)                                               \
+    {                                                                             \
+        if (!(condition__)) {                                                     \
+            std::stringstream _s;                                                 \
+            _s << "Assertion (" <<  #condition__ << ") failed "                   \
+               << "at line " << __LINE__ << " of file " << __FILE__ << std::endl  \
+               << "array label: " << label_ << std::endl;                         \
+            for (int i = 0; i < N; i++) {                                         \
+                _s << "dims[" << i << "].size = " << dims_[i].size() << std::endl;\
+            }                                                                     \
+            throw std::runtime_error(_s.str());                                   \
+            raise(SIGABRT);                                                       \
+        }                                                                         \
+    }
 #endif
 
 /// Index descriptor of mdarray.
@@ -794,7 +795,7 @@ class mdarray
 
     /// Raw pointer.
     T* raw_ptr_{nullptr};
-#ifdef __GPU
+#ifdef SIRIUS_GPU
     /// Unique pointer to the allocated GPU memory.
     std::unique_ptr<T, memory_t_deleter_base> unique_ptr_device_{nullptr};
 
@@ -821,7 +822,7 @@ class mdarray
         }
     }
 
-    /// Return linear index in the range [0, size) by the N-dimensional indices.(i0, i1, ...)
+    /// Return linear index in the range [0, size) by the N-dimensional indices (i0, i1, ...)
     template <typename... Args>
     inline index_type idx(Args... args) const
     {
@@ -850,11 +851,11 @@ class mdarray
                 return &raw_ptr_[idx__];
             }
             case memory_t::device: {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
                 mdarray_assert(raw_ptr_device_ != nullptr);
                 return &raw_ptr_device_[idx__];
 #else
-                printf("error at line %i of file %s: not compiled with GPU support\n", __LINE__, __FILE__);
+                std::printf("error at line %i of file %s: not compiled with GPU support\n", __LINE__, __FILE__);
                 throw std::runtime_error("");
 #endif
             }
@@ -871,9 +872,9 @@ class mdarray
         return const_cast<T*>(static_cast<mdarray<T, N> const&>(*this).at_idx(mem__, idx__));
     }
 
+    // Call constructor on non-trivial data. Complex numbers are treated as trivial.
     inline void call_constructor()
     {
-        /* call constructor on non-trivial data */
         if (!(std::is_trivial<T>::value || is_complex<T>::value)) {
             for (size_t i = 0; i < size(); i++) {
                 new (raw_ptr_ + i) T();
@@ -881,6 +882,7 @@ class mdarray
         }
     }
 
+    // Call destructor on non-trivial data. Complex numbers are treated as trivial.
     inline void call_destructor()
     {
         if (!(std::is_trivial<T>::value || is_complex<T>::value)) {
@@ -898,10 +900,12 @@ class mdarray
 
   public:
 
+    /// Default constructor.
     mdarray()
     {
     }
 
+    /// Destructor.
     ~mdarray()
     {
         deallocate(memory_t::host);
@@ -918,6 +922,11 @@ class mdarray
         this->allocate(memory__);
     }
 
+    /*
+     * 1D array constructors
+     *
+     */
+
     /// 1D array with memory allocation.
     mdarray(mdarray_index_descriptor const& d0,
             memory_t memory__   = memory_t::host,
@@ -930,6 +939,50 @@ class mdarray
         this->allocate(memory__);
     }
 
+    /// 1D array with memory pool allocation.
+    mdarray(mdarray_index_descriptor const& d0, memory_pool& mp__, std::string label__ = "")
+    {
+        static_assert(N == 1, "wrong number of dimensions");
+
+        this->label_ = label__;
+        this->init_dimensions({d0});
+        this->allocate(mp__);
+    }
+
+
+    /// 1D array with host pointer wrapper.
+    mdarray(T* ptr__,
+            mdarray_index_descriptor const& d0,
+            std::string label__ = "")
+    {
+        static_assert(N == 1, "wrong number of dimensions");
+
+        this->label_ = label__;
+        this->init_dimensions({d0});
+        this->raw_ptr_ = ptr__;
+    }
+
+    /// 1D array with host and device pointer wrapper.
+    mdarray(T* ptr__,
+            T* ptr_device__,
+            mdarray_index_descriptor const& d0,
+            std::string label__ = "")
+    {
+        static_assert(N == 1, "wrong number of dimensions");
+
+        this->label_ = label__;
+        this->init_dimensions({d0});
+        this->raw_ptr_ = ptr__;
+#ifdef SIRIUS_GPU
+        this->raw_ptr_device_ = ptr_device__;
+#endif
+    }
+
+    /*
+     * 2D array constructors
+     *
+     */
+
     /// 2D array with memory allocation.
     mdarray(mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
@@ -943,6 +996,7 @@ class mdarray
         this->allocate(memory__);
     }
 
+    /// 3D array with memory allocation.
     mdarray(mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
             mdarray_index_descriptor const& d2,
@@ -956,6 +1010,7 @@ class mdarray
         this->allocate(memory__);
     }
 
+    /// 4D array with memory allocation.
     mdarray(mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
             mdarray_index_descriptor const& d2,
@@ -970,6 +1025,7 @@ class mdarray
         this->allocate(memory__);
     }
 
+    /// 5D array with memory allocation.
     mdarray(mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
             mdarray_index_descriptor const& d2,
@@ -985,6 +1041,7 @@ class mdarray
         this->allocate(memory__);
     }
 
+    /// 6D array with memory allocation.
     mdarray(mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
             mdarray_index_descriptor const& d2,
@@ -1001,43 +1058,6 @@ class mdarray
         this->allocate(memory__);
     }
 
-    mdarray(T* ptr__,
-            mdarray_index_descriptor const& d0,
-            std::string label__ = "")
-    {
-        static_assert(N == 1, "wrong number of dimensions");
-
-        this->label_ = label__;
-        this->init_dimensions({d0});
-        this->raw_ptr_ = ptr__;
-    }
-
-    mdarray(T* ptr__,
-            T* ptr_device__,
-            mdarray_index_descriptor const& d0,
-            std::string label__ = "")
-    {
-        static_assert(N == 1, "wrong number of dimensions");
-
-        this->label_ = label__;
-        this->init_dimensions({d0});
-        this->raw_ptr_ = ptr__;
-#ifdef __GPU
-        this->raw_ptr_device_ = ptr_device__;
-#endif
-    }
-
-    mdarray(memory_pool& mp__,
-            mdarray_index_descriptor const& d0,
-            std::string label__ = "")
-    {
-        static_assert(N == 1, "wrong number of dimensions");
-
-        this->label_ = label__;
-        this->init_dimensions({d0});
-        this->allocate(mp__);
-    }
-
     /// Wrap a pointer into 2D array.
     mdarray(T* ptr__,
             mdarray_index_descriptor const& d0,
@@ -1062,14 +1082,13 @@ class mdarray
         this->label_ = label__;
         this->init_dimensions({d0, d1});
         this->raw_ptr_ = ptr__;
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         this->raw_ptr_device_ = ptr_device__;
 #endif
     }
 
-    mdarray(memory_pool& mp__,
-            mdarray_index_descriptor const& d0,
-            mdarray_index_descriptor const& d1,
+    /// 2D array with memory pool allocation.
+    mdarray(mdarray_index_descriptor const& d0, mdarray_index_descriptor const& d1, memory_pool& mp__,
             std::string label__ = "")
     {
         static_assert(N == 2, "wrong number of dimensions");
@@ -1079,7 +1098,6 @@ class mdarray
         this->allocate(mp__);
     }
 
-
     mdarray(T* ptr__,
             mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
@@ -1105,11 +1123,22 @@ class mdarray
         this->label_ = label__;
         this->init_dimensions({d0, d1, d2});
         this->raw_ptr_ = ptr__;
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         this->raw_ptr_device_ = ptr_device__;
 #endif
     }
 
+    /// 3D array with memory pool allocation.
+    mdarray(mdarray_index_descriptor const& d0, mdarray_index_descriptor const& d1, mdarray_index_descriptor const& d2,
+        memory_pool& mp__, std::string label__ = "")
+    {
+        static_assert(N == 3, "wrong number of dimensions");
+
+        this->label_ = label__;
+        this->init_dimensions({d0, d1, d2});
+        this->allocate(mp__);
+    }
+
     mdarray(T* ptr__,
             mdarray_index_descriptor const& d0,
             mdarray_index_descriptor const& d1,
@@ -1160,7 +1189,7 @@ class mdarray
         : label_(src.label_)
         , unique_ptr_(std::move(src.unique_ptr_))
         , raw_ptr_(src.raw_ptr_)
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         , unique_ptr_device_(std::move(src.unique_ptr_device_))
         , raw_ptr_device_(src.raw_ptr_device_)
 #endif
@@ -1170,12 +1199,12 @@ class mdarray
             offsets_[i] = src.offsets_[i];
         }
         src.raw_ptr_ = nullptr;
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         src.raw_ptr_device_ = nullptr;
 #endif
     }
 
-    /// Move assigment operator
+    /// Move assignment operator
     inline mdarray<T, N>& operator=(mdarray<T, N>&& src)
     {
         if (this != &src) {
@@ -1183,7 +1212,7 @@ class mdarray
             unique_ptr_  = std::move(src.unique_ptr_);
             raw_ptr_     = src.raw_ptr_;
             src.raw_ptr_ = nullptr;
-#ifdef __GPU
+#ifdef SIRIUS_GPU
             unique_ptr_device_  = std::move(src.unique_ptr_device_);
             raw_ptr_device_     = src.raw_ptr_device_;
             src.raw_ptr_device_ = nullptr;
@@ -1210,7 +1239,7 @@ class mdarray
             raw_ptr_    = unique_ptr_.get();
             call_constructor();
         }
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         /* device allocation */
         if (is_device_memory(memory__)) {
             unique_ptr_device_ = get_unique_ptr<T>(this->size(), memory__);
@@ -1233,7 +1262,7 @@ class mdarray
             raw_ptr_    = unique_ptr_.get();
             call_constructor();
         }
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         /* device allocation */
         if (is_device_memory(mp__.memory_type())) {
             unique_ptr_device_ = mp__.get_unique_ptr<T>(this->size());
@@ -1254,7 +1283,7 @@ class mdarray
             unique_ptr_.reset(nullptr);
             raw_ptr_ = nullptr;
         }
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         if (is_device_memory(memory__)) {
             unique_ptr_device_.reset(nullptr);
             raw_ptr_device_ = nullptr;
@@ -1280,7 +1309,7 @@ class mdarray
     /// Access operator[] for the elements of multidimensional array using a linear index in the range [0, size).
     inline T const& operator[](size_t const idx__) const
     {
-        assert(idx__ >= 0 && idx__ < size());
+        mdarray_assert(idx__ >= 0 && idx__ < size());
         return raw_ptr_[idx__];
     }
 
@@ -1349,7 +1378,7 @@ class mdarray
     }
 
     /// Compute hash of the array
-    /** Example: printf("hash(h) : %16llX\n", h.hash()); */
+    /** Example: std::printf("hash(h) : %16llX\n", h.hash()); */
     inline uint64_t hash(uint64_t h__ = 5381) const
     {
         for (size_t i = 0; i < size() * sizeof(T); i++) {
@@ -1384,6 +1413,27 @@ class mdarray
         return checksum(0, size());
     }
 
+    inline T* begin()
+    {
+        return this->at(memory_t::host);
+    }
+
+    inline T const* begin() const
+    {
+        return this->at(memory_t::host);
+    }
+
+    inline T* end()
+    {
+        return this->at(memory_t::host) + this->size();
+    }
+
+    inline T const* end() const
+    {
+        return this->at(memory_t::host) + this->size();
+    }
+
+
     //== template <device_t pu>
     //== inline T checksum() const
     //== {
@@ -1403,12 +1453,19 @@ class mdarray
     //==     }
     //== }
 
-    /// Copy the content of the array to dest
+    /// Copy the content of the array to another array of identical size.
+    /** For example:
+        \code{.cpp}
+        mdarray<double, 2> src(10, 20);
+        mdarray<double, 2> dest(10, 20);
+        src >> dest;
+        \endcode
+     */
     void operator>>(mdarray<T, N>& dest__) const
     {
         for (int i = 0; i < N; i++) {
             if (dest__.dims_[i].begin() != dims_[i].begin() || dest__.dims_[i].end() != dims_[i].end()) {
-                printf("error at line %i of file %s: array dimensions don't match\n", __LINE__, __FILE__);
+                std::printf("error at line %i of file %s: array dimensions don't match\n", __LINE__, __FILE__);
                 raise(SIGTERM);
                 exit(-1);
             }
@@ -1425,7 +1482,7 @@ class mdarray
             //std::fill(raw_ptr_ + idx0__, raw_ptr_ + idx0__ + n__, 0);
             std::memset((void*)&raw_ptr_[idx0__], 0, n__ * sizeof(T));
         }
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         if (n__ && on_device() && is_device_memory(mem__)) {
             mdarray_assert(raw_ptr_device_ != nullptr);
             acc::zero(&raw_ptr_device_[idx0__], n__);
@@ -1442,7 +1499,7 @@ class mdarray
     /// Copy n elements starting from idx0 from one memory type to another.
     inline void copy_to(memory_t mem__, size_t idx0__, size_t n__, stream_id sid = stream_id(-1))
     {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         mdarray_assert(raw_ptr_ != nullptr);
         mdarray_assert(raw_ptr_device_ != nullptr);
         mdarray_assert(idx0__ + n__ <= size());
@@ -1481,7 +1538,7 @@ class mdarray
     /// Check if device pointer is available.
     inline bool on_device() const
     {
-#ifdef __GPU
+#ifdef SIRIUS_GPU
         return (raw_ptr_device_ != nullptr);
 #else
         return false;
@@ -1536,6 +1593,41 @@ std::ostream& operator<<(std::ostream& out, mdarray<T, N>& v)
     return out;
 }
 
+// template for casting matrix with different precision
+template <typename T, typename F, int N>
+inline void copy(mdarray<F, N> const& src__, mdarray<T, N>& dest__)
+{
+    if (src__.size() == 0) {
+        return;
+    }
+    for (int i = 0; i < N; i++) {
+        if (dest__.dim(i).begin() != src__.dim(i).begin() || dest__.dim(i).end() != src__.dim(i).end()) {
+            std::stringstream s;
+            s << "error at line " << __LINE__ << " of file " << __FILE__ << " : array dimensions don't match";
+            throw std::runtime_error(s.str());
+        }
+    }
+    std::cout << "=== WARNING at line " << __LINE__ << " of file " << __FILE__ << " ===" << std::endl;
+    std::cout << "    Copying matrix element with different type, possible lost of data precision" << std::endl;
+    std::copy(&src__.at(memory_t::host)[0], &src__.at(memory_t::host)[0] + src__.size(), &dest__.at(memory_t::host)[0]);
+}
+
+template <typename T, int N>
+inline void copy(mdarray<T, N> const& src__, mdarray<T, N>& dest__)
+{
+    if (src__.size() == 0) {
+        return;
+    }
+    for (int i = 0; i < N; i++) {
+        if (dest__.dim(i).begin() != src__.dim(i).begin() || dest__.dim(i).end() != src__.dim(i).end()) {
+            std::stringstream s;
+            s << "error at line " << __LINE__ << " of file " << __FILE__ << " : array dimensions don't match";
+            throw std::runtime_error(s.str());
+        }
+    }
+    std::copy(&src__[0], &src__[0] + src__.size(), &dest__[0]);
+}
+
 }
 
 #endif  // __MEMORY_HPP__
