@@ -27,18 +27,22 @@
 
 #include <cstdlib>
 #include <iostream>
-#include <omp.h>
-#include "linalg.hpp"
-#include "eigenproblem.hpp"
-#include "hdf5_tree.hpp"
+#include "linalg/linalg.hpp"
+#include "SDDK/hdf5_tree.hpp"
 #include "utils/env.hpp"
-#include "gvec.hpp"
+#include "SDDK/gvec.hpp"
 #include "matrix_storage.hpp"
-#ifdef __GPU
-extern "C" void add_square_sum_gpu(double_complex const* wf__, int num_rows_loc__, int nwf__, int reduced__,
+#include "type_definition.hpp"
+#ifdef SIRIUS_GPU
+extern "C" void add_square_sum_gpu_double(double_complex const* wf__, int num_rows_loc__, int nwf__, int reduced__,
                                    int mpi_rank__, double* result__);
 
-extern "C" void add_checksum_gpu(double_complex* wf__, int num_rows_loc__, int nwf__, double_complex* result__);
+extern "C" void add_square_sum_gpu_float(std::complex<float> const* wf__, int num_rows_loc__, int nwf__, int reduced__,
+                                   int mpi_rank__, float* result__);
+
+extern "C" void scale_matrix_columns_gpu_double(int nrow__, int ncol__, std::complex<double>* mtrx__, double* a__);
+
+extern "C" void scale_matrix_columns_gpu_float(int nrow__, int ncol__, std::complex<float>* mtrx__, float* a__);
 #endif
 
 const int sddk_inner_default_block_size = 1024;
@@ -46,20 +50,19 @@ const int sddk_trans_default_block_size = 2048;
 
 namespace sddk {
 
-/// Helper function: get a list of spin-indices.
-inline std::vector<int> get_spins(int ispn__)
-{
-    return (ispn__ == 2) ? std::vector<int>({0, 1}) : std::vector<int>({ispn__});
-}
+// C++ wrappers for gpu kernels
+void add_square_sum_gpu(std::complex<double> const* wf__, int num_rows_loc__, int nwf__, int reduced__, int mpi_rank__, double* result__);
+void add_square_sum_gpu(std::complex<float> const* wf__, int num_rows_loc__, int nwf__, int reduced__, int mpi_rank__, float* result__);
+void scale_matrix_columns_gpu(int nrow__, int ncol__, std::complex<double>* mtrx__, double* a__);
+void scale_matrix_columns_gpu(int nrow__, int ncol__, std::complex<float>* mtrx__, float* a__);
 
 /// Helper class to wrap spin index range.
 /** Depending on the collinear or non-collinear case, the spin index range of the wave-functions is either
  *  [0, 0] or [1, 1] (trivial cases of single spin channel) or [0, 1] (spinor wave-functions). */
-class spin_range
+class spin_range : public std::vector<int>
 {
   private:
     int idx_;
-    std::vector<int> v_;
   public:
     explicit spin_range(int idx__)
         : idx_(idx__)
@@ -68,27 +71,22 @@ class spin_range
             throw std::runtime_error("wrong spin index");
         }
         if (idx_ == 2) {
-            v_ = std::vector<int>({0, 1});
+            this->reserve(2);
+            this->push_back(0);
+            this->push_back(1);
         } else {
-            v_ = std::vector<int>({idx_});
+            this->reserve(1);
+            this->push_back(idx_);
         }
     }
     inline int operator()() const
     {
         return idx_;
     }
-
-    inline std::vector<int>::const_iterator begin() const
-    {
-        return v_.begin();
-    }
-
-    inline std::vector<int>::const_iterator end() const
-    {
-        return v_.end();
-    }
 };
 
+//TODO introduce band range to desctibe a set of bands in the interval [N1, N2]
+
 /// Wave-functions representation.
 /** Wave-functions consist of two parts: plane-wave part and mufin-tin part. Both are the matrix_storage objects
  *  with the slab distribution. Wave-functions have one or two spin components. In case of collinear magnetism
@@ -164,6 +162,7 @@ class spin_range
     }
     \endcode
  */
+template <typename T>   // template type is using real type that determine the precision, wavefunction is always complex
 class Wave_functions
 {
   private:
@@ -190,11 +189,11 @@ class Wave_functions
     int num_sc_{1};
 
     /// Plane-wave part of wave-functions.
-    std::array<std::unique_ptr<matrix_storage<double_complex, matrix_storage_t::slab>>, 2> pw_coeffs_{
+    std::array<std::unique_ptr<matrix_storage<std::complex<T>, matrix_storage_t::slab>>, 2> pw_coeffs_{
         {nullptr, nullptr}};
 
     /// Muffin-tin part of wave-functions.
-    std::array<std::unique_ptr<matrix_storage<double_complex, matrix_storage_t::slab>>, 2> mt_coeffs_{
+    std::array<std::unique_ptr<matrix_storage<std::complex<T>, matrix_storage_t::slab>>, 2> mt_coeffs_{
         {nullptr, nullptr}};
 
     bool has_mt_{false};
@@ -235,13 +234,15 @@ class Wave_functions
 
     /// Compute the sum of squares of expansion coefficients.
     /** The result is always returned in the host memory */
-    mdarray<double, 1> sumsqr(device_t pu__, spin_range spins__, int n__) const;
+    mdarray<T, 1> sumsqr(device_t pu__, spin_range spins__, int n__) const;
 
   public:
     /// Constructor for PW wave-functions.
+    /** Memory to store plane-wave coefficients is allocated from the heap. */
     Wave_functions(Gvec_partition const& gkvecp__, int num_wf__, memory_t preferred_memory_t__, int num_sc__ = 1);
 
     /// Constructor for PW wave-functions.
+    /** Memory to store plane-wave coefficients is allocated from the memory pool. */
     Wave_functions(memory_pool& mp__, Gvec_partition const& gkvecp__, int num_wf__, memory_t preferred_memory_t__,
                    int num_sc__ = 1);
 
@@ -271,22 +272,22 @@ class Wave_functions
         return num_mt_coeffs_;
     }
 
-    inline matrix_storage<double_complex, matrix_storage_t::slab>& pw_coeffs(int ispn__)
+    inline matrix_storage<std::complex<T>, matrix_storage_t::slab>& pw_coeffs(int ispn__)
     {
         return *pw_coeffs_[isc(ispn__)];
     }
 
-    inline matrix_storage<double_complex, matrix_storage_t::slab> const& pw_coeffs(int ispn__) const
+    inline matrix_storage<std::complex<T>, matrix_storage_t::slab> const& pw_coeffs(int ispn__) const
     {
         return *pw_coeffs_[isc(ispn__)];
     }
 
-    inline matrix_storage<double_complex, matrix_storage_t::slab>& mt_coeffs(int ispn__)
+    inline matrix_storage<std::complex<T>, matrix_storage_t::slab>& mt_coeffs(int ispn__)
     {
         return *mt_coeffs_[isc(ispn__)];
     }
 
-    inline matrix_storage<double_complex, matrix_storage_t::slab> const& mt_coeffs(int ispn__) const
+    inline matrix_storage<std::complex<T>, matrix_storage_t::slab> const& mt_coeffs(int ispn__) const
     {
         return *mt_coeffs_[isc(ispn__)];
     }
@@ -316,6 +317,22 @@ class Wave_functions
         return offset_mt_coeffs_[ialoc__];
     }
 
+    inline memory_t preferred_memory_t() const
+    {
+        return preferred_memory_t_;
+    }
+
+    inline auto checksum(device_t pu__, int ispn__, int i0__, int n__) const
+    {
+        return checksum_pw(pu__, ispn__, i0__, n__) + checksum_mt(pu__, ispn__, i0__, n__);
+    }
+
+    inline void zero(device_t pu__, int ispn__, int i0__, int n__) // TODO: pass memory_t
+    {
+        zero_pw(pu__, ispn__, i0__, n__);
+        zero_mt(pu__, ispn__, i0__, n__);
+    }
+
     /// Copy values from another wave-function.
     /** \param [in] pu   Type of processging unit which copies data.
      *  \param [in] n    Number of wave-functions to copy.
@@ -324,50 +341,116 @@ class Wave_functions
      *  \param [in] i0   Starting index of wave-functions in src.
      *  \param [in] jspn Spin component on destination wave-functions.
      *  \param [in] j0   Starting index of wave-functions in destination. */
-    void copy_from(device_t pu__, int n__, Wave_functions const& src__, int ispn__, int i0__, int jspn__,
-                          int j0__);
+    void copy_from(device_t pu__, int n__, Wave_functions<T> const& src__, int ispn__, int i0__, int jspn__, int j0__);
 
-    void copy_from(Wave_functions const& src__, int n__, int ispn__, int i0__, int jspn__, int j0__);
+    template <typename F>
+    void copy_from(device_t pu__, int n__, Wave_functions<F> const& src__, int ispn__, int i0__, int jspn__, int j0__) {
+        assert(ispn__ == 0 || ispn__ == 1);
+        assert(jspn__ == 0 || jspn__ == 1);
+        std::cout << "=== WARNING at line " << __LINE__ << " of file " << __FILE__ << " ===" << std::endl;
+        std::cout << "    Copying Wavefunction with different type, possible lost of data precision" << std::endl;
+
+        int ngv = pw_coeffs(jspn__).num_rows_loc();
+        int nmt = has_mt() ? mt_coeffs(jspn__).num_rows_loc() : 0;
+
+        switch (pu__) {
+            case device_t::CPU: {
+                /* copy PW part */
+                std::copy(src__.pw_coeffs(ispn__).prime().at(memory_t::host, 0, i0__),
+                          src__.pw_coeffs(ispn__).prime().at(memory_t::host, 0, i0__) + ngv * n__,
+                          pw_coeffs(jspn__).prime().at(memory_t::host, 0, j0__));
+                /* copy MT part */
+                if (has_mt()) {
+                    std::copy(src__.mt_coeffs(ispn__).prime().at(memory_t::host, 0, i0__),
+                              src__.mt_coeffs(ispn__).prime().at(memory_t::host, 0, i0__) + nmt * n__,
+                              mt_coeffs(jspn__).prime().at(memory_t::host, 0, j0__));
+                }
+                break;
+            }
+            case device_t::GPU: {
+                throw std::runtime_error("Copy mixed precision type not supported in device memory");
+                break;
+            }
+        }
+    }
+
+    /// Copy from and to preferred memory.
+    void copy_from(Wave_functions<T> const& src__, int n__, int ispn__, int i0__, int jspn__, int j0__);
+
+    template <typename F>
+    void copy_from(Wave_functions<F> const& src__, int n__, int ispn__, int i0__, int jspn__, int j0__) {
+        assert(ispn__ == 0 || ispn__ == 1);
+        assert(jspn__ == 0 || jspn__ == 1);
+        std::cout << "=== WARNING at line " << __LINE__ << " of file " << __FILE__ << " ===" << std::endl;
+        std::cout << "    Copying Wavefunction with different type, possible lost of data precision" << std::endl;
+
+        int ngv = pw_coeffs(jspn__).num_rows_loc();
+        int nmt = has_mt() ? mt_coeffs(jspn__).num_rows_loc() : 0;
+
+        copy(src__.preferred_memory_t(), src__.pw_coeffs(ispn__).prime().at(src__.preferred_memory_t(), 0, i0__),
+             preferred_memory_t(), pw_coeffs(jspn__).prime().at(preferred_memory_t(), 0, j0__), ngv * n__);
+        if (has_mt()) {
+            copy(src__.preferred_memory_t(), src__.mt_coeffs(ispn__).prime().at(src__.preferred_memory_t(), 0, i0__),
+                 preferred_memory_t(), mt_coeffs(jspn__).prime().at(preferred_memory_t(), 0, j0__), nmt * n__);
+        }
+    }
 
     /// Compute the checksum of the spin-components.
     /** Checksum of the n wave-function spin components is computed starting from i0.
      *  Only plane-wave coefficients are considered. */
-    double_complex checksum_pw(device_t pu__, int ispn__, int i0__, int n__);
+    std::complex<T> checksum_pw(device_t pu__, int ispn__, int i0__, int n__) const;
 
     /// Checksum of muffin-tin coefficients.
-    double_complex checksum_mt(device_t pu__, int ispn__, int i0__, int n__);
-
-    inline double_complex checksum(device_t pu__, int ispn__, int i0__, int n__)
-    {
-        return checksum_pw(pu__, ispn__, i0__, n__) + checksum_mt(pu__, ispn__, i0__, n__);
-    }
+    std::complex<T> checksum_mt(device_t pu__, int ispn__, int i0__, int n__) const;
 
     void zero_pw(device_t pu__, int ispn__, int i0__, int n__);
 
     void zero_mt(device_t pu__, int ispn__, int i0__, int n__);
 
-    inline void zero(device_t pu__, int ispn__, int i0__, int n__) // TODO: pass memory_t
-    {
-        zero_pw(pu__, ispn__, i0__, n__);
-        zero_mt(pu__, ispn__, i0__, n__);
-    }
+    void scale(memory_t mem__, int ispn__, int i0__, int n__, T beta__);
 
-    void scale(memory_t mem__, int ispn__, int i0__, int n__, double beta__);
-
-    mdarray<double, 1> l2norm(device_t pu__, spin_range spins__, int n__) const;
+    sddk::mdarray<T, 1> l2norm(device_t pu__, spin_range spins__, int n__) const;
 
     /// Normalize the functions.
     void normalize(device_t pu__, spin_range spins__, int n__);
 
     void allocate(spin_range spins__, memory_t mem__);
 
+    void allocate(spin_range spins__, memory_pool& mp__);
+
     void deallocate(spin_range spins__, memory_t mem__);
 
     void copy_to(spin_range spins__, memory_t mem__, int i0__, int n__);
 
-    inline memory_t preferred_memory_t() const
+    void print_checksum(device_t pu__, std::string label__, int N__, int n__) const;
+
+    /// Prepare wave-functions on the device.
+    void prepare(spin_range spins__, bool with_copy__, memory_pool* mp__ = nullptr)
     {
-        return preferred_memory_t_;
+        /* if operations on wave-functions are done on GPUs */
+        if (is_device_memory(preferred_memory_t_)) {
+            if (mp__) {
+                if (!is_device_memory(mp__->memory_type())) {
+                    TERMINATE("not a device memory pool");
+                }
+                this->allocate(spins__, *mp__);
+            } else {
+                this->allocate(spins__, preferred_memory_t_);
+            }
+            if (with_copy__) {
+                this->copy_to(spins__, preferred_memory_t_, 0, this->num_wf());
+            }
+        }
+    }
+
+    void dismiss(spin_range spins__, bool with_copy__)
+    {
+        if (is_device_memory(preferred_memory_t_)) {
+            if (with_copy__) {
+                this->copy_to(spins__, memory_t::host, 0, this->num_wf());
+            }
+            this->deallocate(spins__, preferred_memory_t_);
+        }
     }
 };
 
