@@ -31,6 +31,8 @@
 #include <vector>
 #include <array>
 #include <limits>
+#include <map>
+#include <algorithm>
 
 using double_complex = std::complex<double>;
 
@@ -72,6 +74,7 @@ enum class function_domain_t
     spectral
 };
 
+/// Type of relativity treatment in the case of LAPW.
 enum class relativity_t
 {
     none,
@@ -85,6 +88,25 @@ enum class relativity_t
     dirac
 };
 
+inline relativity_t get_relativity_t(std::string name__)
+{
+    std::transform(name__.begin(), name__.end(), name__.begin(), ::tolower);
+    std::map<std::string, relativity_t> const m = {
+        {"none", relativity_t::none},
+        {"koelling_harmon", relativity_t::koelling_harmon},
+        {"zora", relativity_t::zora},
+        {"iora", relativity_t::iora},
+        {"dirac", relativity_t::dirac}
+    };
+
+    if (m.count(name__) == 0) {
+        std::stringstream s;
+        s << "get_relativity_t(): wrong label of the relativity_t enumerator: " << name__;
+        throw std::runtime_error(s.str());
+     }
+     return m.at(name__);
+}
+
 /// Describes radial solution.
 struct radial_solution_descriptor
 {
@@ -116,7 +138,7 @@ struct local_orbital_descriptor
     int l;
 
     /// Total angular momentum used in pseudopotential SO code.
-    double total_angular_momentum; // TODO: is this neccessary?
+    double total_angular_momentum; // TODO: is this necessary?
 
     /// Set of radial solution descriptors.
     /** Local orbital is constructed from at least two radial functions in order to make it zero at the
@@ -124,7 +146,7 @@ struct local_orbital_descriptor
     radial_solution_descriptor_set rsd_set;
 };
 
-/// Descriptor of an atom in a list of nearest neigbours for each atom.
+/// Descriptor of an atom in a list of nearest neighbours for each atom.
 /** See sirius::Unit_cell::find_nearest_neighbours() for the details of usage. */
 struct nearest_neighbour_descriptor
 {
@@ -136,6 +158,9 @@ struct nearest_neighbour_descriptor
 
     /// Distance from the central atom.
     double distance;
+
+    /// Vector connecting central atom with the neighbour in Cartesian coordinates.
+    std::array<double, 3> rc;
 };
 
 struct unit_cell_parameters_descriptor
